import inspect

import numpy as np
from typing_extensions import List, TYPE_CHECKING
from scipy.spatial.transform import Rotation as R

from pycrap.urdf_parser import parse_furniture
from ..datastructures.dataclasses import Colors
from ..datastructures.world import World
from ..designators.motion_designator import *
from ..external_interfaces import giskard
from ..external_interfaces.ik import request_ik
from ..external_interfaces.move_base import query_pose_nav
from ..external_interfaces.robokudo import query_all_objects, query_object, query_human, query_specific_region, \
    query_human_attributes, query_waving_human
from ..failures import NavigationGoalNotReachedError
from ..local_transformer import LocalTransformer
from ..object_descriptors.generic import ObjectDescription as GenericObjectDescription
from ..process_module import ProcessModule
from ..robot_description import RobotDescription
from ..ros import get_time
from ..ros import logdebug, loginfo
from ..tf_transformations import euler_from_quaternion
from ..utils import _apply_ik
from ..world_concepts.world_object import Object
from ..world_reasoning import visible, link_pose_for_joint_config

if TYPE_CHECKING:
    from ..designators.object_designator import ObjectDesignatorDescription


class DefaultNavigation(ProcessModule):
    """
    The process module to move the robot from one position to another.
    """

    def _execute(self, desig: MoveMotion):
        robot = World.robot
        robot.set_pose(desig.target)


class DefaultMoveHead(ProcessModule):
    """
    Moves the robot's head to look at a specified target point in the world coordinate frame.
    The target can be either a position or an object.
    """

    def _execute(self, desig: LookingMotion):
        target = desig.target
        robot = World.robot
        local_transformer = LocalTransformer()

        neck = RobotDescription.current_robot_description.get_neck()
        pan_link = neck["yaw"][0]
        tilt_link = neck["pitch"][0]

        pan_joint = neck["yaw"][1]
        tilt_joint = neck["pitch"][1]

        pose_in_map = local_transformer.transform_pose(target, "map")

        pose_in_pan = local_transformer.transform_pose(pose_in_map,
                                                       robot.get_link_tf_frame(pan_link)).position.to_list()
        pose_in_tilt = local_transformer.transform_pose(pose_in_map,
                                                        robot.get_link_tf_frame(tilt_link)).position.to_list()

        new_pan = np.arctan2(pose_in_pan[1], pose_in_pan[0])

        tilt_offset = RobotDescription.current_robot_description.get_offset(tilt_joint)
        if tilt_offset:
            tilt_offset_rotation = tilt_offset.orientation
            quaternion_list = [tilt_offset_rotation.x, tilt_offset_rotation.y, tilt_offset_rotation.z,
                               tilt_offset_rotation.w]
        else:
            quaternion_list = [0, 0, 0, 1]

        tilt_offset_rotation = euler_from_quaternion(quaternion_list, axes='sxyz')
        adjusted_pose_in_tilt = R.from_euler('xyz', tilt_offset_rotation).apply(pose_in_tilt)

        new_tilt = -np.arctan2(adjusted_pose_in_tilt[2],
                               np.sqrt(adjusted_pose_in_tilt[0] ** 2 + adjusted_pose_in_tilt[1] ** 2))

        if RobotDescription.current_robot_description.name in {"iCub", "tiago_dual"}:
            new_tilt = -new_tilt

        current_pan = robot.get_joint_position(pan_joint)
        current_tilt = robot.get_joint_position(tilt_joint)
        robot.set_joint_position(pan_joint, new_pan + current_pan)
        robot.set_joint_position(tilt_joint, new_tilt + current_tilt)


class DefaultMoveGripper(ProcessModule):
    """
    This process module controls the gripper of the robot. They can either be opened or closed.
    Furthermore, it can only move one gripper at a time.
    """

    def _execute(self, desig: MoveGripperMotion):
        robot_description = RobotDescription.current_robot_description
        gripper = desig.gripper
        arm_chain = robot_description.get_arm_chain(gripper)
        if arm_chain.end_effector.gripper_object_name is not None:
            robot = World.current_world.get_object_by_name(arm_chain.end_effector.gripper_object_name)
        else:
            robot = World.robot
        motion = desig.motion
        robot.set_multiple_joint_positions(arm_chain.get_static_gripper_state(motion))


class DefaultDetecting(ProcessModule):
    """
    This process module tries to detect an object with the given type. To be detected the object has to be in
    the field of view of the robot.
    :return: A list of perceived objects.
    """

    def _execute(self, designator: DetectingMotion):
        robot = World.robot
        cam_link_name = RobotDescription.current_robot_description.get_camera_link()
        camera_description = RobotDescription.current_robot_description.cameras[
            list(RobotDescription.current_robot_description.cameras.keys())[0]]
        front_facing_axis = camera_description.front_facing_axis
        query_result = []
        world_objects = []

        if designator.technique == DetectionTechnique.TYPES:
            try:
                object_types = designator.object_designator_description.obj_type
            except AttributeError:
                raise AttributeError("The object designator does not contain a type attribute")

            list1 = World.current_world.get_object_by_type(object_types)
            world_objects = world_objects + list1
        elif designator.technique == DetectionTechnique.ALL:
            world_objects = World.current_world.get_scene_objects()
        elif designator.technique == DetectionTechnique.HUMAN:
            raise NotImplementedError("Detection by human is not yet implemented in simulation")
        elif designator.technique == DetectionTechnique.REGION:
            raise NotImplementedError("Detection by region is not yet implemented in simulation")
        elif designator.technique == DetectionTechnique.HUMAN_ATTRIBUTES:
            raise NotImplementedError("Detection by human attributes is not yet implemented in simulation")
        elif designator.technique == DetectionTechnique.HUMAN_WAVING:
            raise NotImplementedError("Detection by waving human is not yet implemented in simulation")
        for obj in world_objects:
            if visible(obj, robot.get_link_pose(cam_link_name), front_facing_axis):
                query_result.append(obj)
        if query_result is None:
            raise PerceptionObjectNotFound(
                f"Could not find an object with the type {object_types} in the FOV of the robot")
        else:
            object_dict = []

            for obj in query_result:
                object_dict.append(obj)

            return object_dict


class DefaultMoveTCP(ProcessModule):
    """
    This process moves the tool center point of either the right or the left arm.
    """

    def _execute(self, desig: MoveTCPMotion):
        target = desig.target
        robot = World.robot

        _move_arm_tcp(target, robot, desig.arm)


class DefaultMoveArmJoints(ProcessModule):
    """
    This process modules moves the joints of either the right or the left arm. The joint states can be given as
    list that should be applied or a pre-defined position can be used, such as "parking"
    """

    def _execute(self, desig: MoveArmJointsMotion):

        robot = World.robot
        if desig.right_arm_poses:
            for joint, pose in desig.right_arm_poses.items():
                robot.set_joint_position(joint, pose)
        if desig.left_arm_poses:
            for joint, pose in desig.left_arm_poses.items():
                robot.set_joint_position(joint, pose)


class DefaultMoveJoints(ProcessModule):
    def _execute(self, desig: MoveJointsMotion):
        robot = World.robot
        for joint, pose in zip(desig.names, desig.positions):
            robot.set_joint_position(joint, pose)


class DefaultWorldStateDetecting(ProcessModule):
    """
    This process moduledetectes an object even if it is not in the field of view of the robot.
    """

    def _execute(self, desig: WorldStateDetectingMotion):
        obj_type = desig.object_type
        return list(filter(lambda obj: obj.obj_type == obj_type, World.current_world.objects))[0]


class DefaultOpen(ProcessModule):
    """
    Low-level implementation of opening a container in the simulation. Assumes the handle is already grasped.
    """

    def _execute(self, desig: OpeningMotion):
        part_of_object = desig.object_part.parent_entity

        container_joint_name = part_of_object.find_joint_above_link(desig.object_part.name)
        lower_limit, upper_limit = part_of_object.get_joint_limits(container_joint_name)

        goal_pose = link_pose_for_joint_config(part_of_object, {
            container_joint_name: max(lower_limit, upper_limit - 0.05)}, desig.object_part.name)

        _move_arm_tcp(goal_pose, World.robot, desig.arm)

        part_of_object.set_joint_position(container_joint_name, upper_limit)


class DefaultClose(ProcessModule):
    """
    Low-level implementation that lets the robot close a grasped container, in simulation
    """

    def _execute(self, desig: ClosingMotion):
        part_of_object = desig.object_part.parent_entity

        container_joint_name = part_of_object.find_joint_above_link(desig.object_part.name)
        lower_joint_limit = part_of_object.get_joint_limits(container_joint_name)[0]

        goal_pose = link_pose_for_joint_config(part_of_object, {
            container_joint_name: lower_joint_limit}, desig.object_part.name)

        _move_arm_tcp(goal_pose, World.robot, desig.arm)

        part_of_object.set_joint_position(container_joint_name, lower_joint_limit)


class DefaultMoveTCPWaypoints(ProcessModule):
    """
    This process moves the tool center point of either the right or the left arm along a list of waypoints.
    """

    def _execute(self, desig: MoveTCPWaypointsMotion):
        waypoints = desig.waypoints
        robot = World.robot
        for waypoint in waypoints:
            _move_arm_tcp(waypoint, robot, desig.arm)


def _move_arm_tcp(target: PoseStamped, robot: Object, arm: Arms, tip_link: str = None) -> None:
    """
    Calls the ik solver to calculate the inverse kinematics of the arm and then sets the joint states accordingly.

    :param target: Target pose to which the end-effector should move.
    :param robot: Robot object representing the robot.
    :param arm: Which arm to move
    """
    if tip_link is None:
        tip_link = RobotDescription.current_robot_description.get_arm_chain(arm).get_tool_frame()

    joints = RobotDescription.current_robot_description.get_arm_chain(arm).joints

    inv = request_ik(target, robot, joints, tip_link)
    _apply_ik(robot, inv)


###########################################################
########## Process Modules for the Real     ###############
###########################################################

class DefaultDetectingReal(ProcessModule):
    def _execute(self, designator: DetectingMotion) -> List[Object]:
        """
            Perform a query based on the detection technique and state defined in the designator.

            :return: A list of perceived objects.
            """
        print(designator.technique)
        object_designator_description = designator.object_designator_description
        query_methods = {
            DetectionTechnique.TYPES: lambda: query_object(object_designator_description),
            DetectionTechnique.HUMAN: lambda: query_human(),
            DetectionTechnique.HUMAN_ATTRIBUTES: query_human_attributes,
            DetectionTechnique.HUMAN_WAVING: query_waving_human,
            DetectionTechnique.REGION: lambda: query_specific_region(designator.region)
        }  # Fetch the appropriate query function
        query_func = query_methods.get(designator.technique, query_all_objects)
        query_result = query_func() if callable(query_func) else query_func
        # Handle the case where no result is found
        if query_result is None:
            raise PerceptionObjectNotFound(
                f"Could not find an object in the FOV of the robot")
        else:
            perceived_objects = []
            for i in range(0, len(query_result.res)):
                try:
                    obj_pose = PoseStamped.from_ros_message(query_result.res[i].pose[0])
                except IndexError:
                    obj_pose = PoseStamped.from_ros_message(query_result.res[i].pose)
                    pass
                lt = LocalTransformer()
                obj_pose.frame_id = World.robot.get_link_tf_frame(obj_pose.frame_id)
                obj_pose_T_m = lt.transform_pose(obj_pose, "map")

                obj_type = query_result.res[i].type
                obj_size = None
                try:
                    obj_size = query_result.res[i].shape_size[0].dimensions
                except IndexError:
                    pass

                obj_color = None
                try:
                    obj_color = query_result.res[i].color[0]
                except IndexError:
                    pass

                if obj_size is None:
                    hsize = [0.2, 0.2, 0.2]

                else:
                    hsize = [obj_size.x / 2, obj_size.y / 2, obj_size.z / 2]

                # Check if the object type is a subclass of the classes in the objects module (pycrap)
                type_concept = parse_furniture(obj_type)
                if type_concept is None:
                    loginfo(f"No class name contains the string '{obj_type}'")
                    type_concept = PhysicalObject
                print(type_concept)
                obj_name = obj_type + "" + str(get_time())
                print(obj_name)
                gen_obj_desc = GenericObjectDescription(obj_name, [0, 0, 0], hsize)
                print("1")
                if obj_color is not None: color = Colors.from_string(obj_color)
                else: color = Colors.PINK
                print("2")
                generic_obj = Object(name=obj_name, concept=type_concept, path=None, description=gen_obj_desc,
                                     color=color)
                print("3")

                generic_obj.set_pose(obj_pose)
                print("4")

                perceived_objects.append(generic_obj)

            object_dict = []

            for obj in perceived_objects:
                object_dict.append(ObjectDesignatorDescription.Object(obj.name, obj.obj_type,
                                                                      obj))

            return object_dict


class DefaultNavigationReal(ProcessModule):
    """
    Process module for the real robot that sends a cartesian goal to giskard to move the robot base
    """

    def _execute(self, designator: MoveMotion):
        # logdebug(f"Sending goal to movebase to Move the robot")
        # query_pose_nav(designator.target)
        logdebug(f"Sending goal to giskard to Move the robot")
        giskard.avoid_all_collisions()
        giskard.achieve_cartesian_goal(designator.target,
                                       RobotDescription.current_robot_description.base_link,
                                       "map")

        if not World.current_world.robot.pose.almost_equal(designator.target, 0.05, 3):
            raise NavigationGoalNotReachedError(World.current_world.robot.pose, designator.target)


class DefaultMoveHeadReal(ProcessModule):
    """
    Process module for controlling the real robot's head to look at a specified position.
    Uses the same calculations as the simulated version to orient the head.
    """

    def _execute(self, desig: LookingMotion):
        target = desig.target
        robot = World.robot

        local_transformer = LocalTransformer()
        neck = RobotDescription.current_robot_description.get_neck()
        pan_link = neck["yaw"][0]
        tilt_link = neck["pitch"][0]

        pan_joint = neck["yaw"][1]
        tilt_joint = neck["pitch"][1]

        pose_in_pan = local_transformer.transform_pose(target, robot.get_link_tf_frame(pan_link)).position.to_list()
        pose_in_tilt = local_transformer.transform_pose(target, robot.get_link_tf_frame(tilt_link)).position.to_list()

        new_pan = np.arctan2(pose_in_pan[1], pose_in_pan[0])

        tilt_offset = RobotDescription.current_robot_description.get_offset(tilt_joint)
        if tilt_offset:
            tilt_offset_rotation = tilt_offset.pose.orientation
            quaternion_list = [tilt_offset_rotation.x, tilt_offset_rotation.y, tilt_offset_rotation.z,
                               tilt_offset_rotation.w]
        else:
            quaternion_list = [0, 0, 0, 1]

        tilt_offset_rotation = euler_from_quaternion(quaternion_list, axes='sxyz')
        adjusted_pose_in_tilt = R.from_euler('xyz', tilt_offset_rotation).apply(pose_in_tilt)

        new_tilt = -np.arctan2(adjusted_pose_in_tilt[2],
                               np.sqrt(adjusted_pose_in_tilt[0] ** 2 + adjusted_pose_in_tilt[1] ** 2))

        if RobotDescription.current_robot_description.name in {"iCub", "tiago_dual"}:
            new_tilt = -new_tilt

        current_pan = robot.get_joint_position(pan_joint)
        current_tilt = robot.get_joint_position(tilt_joint)

        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal({pan_joint: new_pan + current_pan,
                                    tilt_joint: new_tilt + current_tilt})


class DefaultMoveTCPReal(ProcessModule):
    """
    Moves the tool center point of the real robot while avoiding all collisions
    """

    def _execute(self, designator: MoveTCPMotion):
        lt = LocalTransformer()
        pose_in_map = lt.transform_pose(designator.target, "map")
        tip_link = RobotDescription.current_robot_description.get_arm_chain(designator.arm).get_tool_frame()
        root_link = "map"

        gripper_that_can_collide = designator.arm if designator.allow_gripper_collision else None
        if designator.allow_gripper_collision:
            giskard.allow_gripper_collision(designator.arm.name.lower())

        if designator.movement_type == MovementType.STRAIGHT_TRANSLATION:
            giskard.achieve_straight_translation_goal(pose_in_map.position.to_list(), tip_link, root_link)
        elif designator.movement_type == MovementType.STRAIGHT_CARTESIAN:
            giskard.achieve_straight_cartesian_goal(pose_in_map, tip_link, root_link)
        elif designator.movement_type == MovementType.TRANSLATION:
            giskard.achieve_translation_goal(pose_in_map.position.to_list(), tip_link, root_link)
        elif designator.movement_type == MovementType.CARTESIAN:
            print("i think i spider")
            giskard.achieve_cartesian_goal(pose_in_map, tip_link, root_link,
                                           grippers_that_can_collide=gripper_that_can_collide,
                                          )
        if not World.current_world.robot.get_link_pose(tip_link).almost_equal(designator.target, 0.01, 3):
            raise ToolPoseNotReachedError(World.current_world.robot.get_link_pose(tip_link), designator.target)


class DefaultMoveArmJointsReal(ProcessModule):
    """
    Moves the arm joints of the real robot to the given configuration while avoiding all collisions
    """

    def _execute(self, designator: MoveArmJointsMotion):
        joint_goals = {}
        if designator.left_arm_poses:
            joint_goals.update(designator.left_arm_poses)
        if designator.right_arm_poses:
            joint_goals.update(designator.right_arm_poses)
        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal(joint_goals)


class DefaultMoveJointsReal(ProcessModule):
    """
    Moves any joint using giskard, avoids all collisions while doint this.
    """

    def _execute(self, designator: MoveJointsMotion):
        name_to_position = dict(zip(designator.names, designator.positions))
        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal(name_to_position)


class DefaultMoveGripperReal(ProcessModule):
    """
    Opens or closes the gripper of the real robot, gripper uses an action server for this instead of giskard
    """

    def _execute(self, designator: MoveGripperMotion):
        raise NotImplementedError(f"There is DefaultMoveGripperReal process module")


class DefaultOpenReal(ProcessModule):
    """
    Tries to open an already grasped container
    """

    def _execute(self, designator: OpeningMotion):
        giskard.achieve_open_container_goal(
            RobotDescription.current_robot_description.get_arm_chain(designator.arm).get_tool_frame(),
            designator.object_part.name)


class DefaultCloseReal(ProcessModule):
    """
    Tries to close an already grasped container
    """

    def _execute(self, designator: ClosingMotion):
        giskard.achieve_close_container_goal(
            RobotDescription.current_robot_description.get_arm_chain(designator.arm).get_tool_frame(),
            designator.object_part.name)


class DefaultMoveTCPWaypointsReal(ProcessModule):
    """
    Moves the tool center point of the real robot along a list of waypoints while avoiding all collisions
    """

    def _execute(self, designator: MoveTCPWaypointsMotion):
        lt = LocalTransformer()
        waypoints = [lt.transform_pose(x, "map") for x in designator.waypoints]
        tip_link = RobotDescription.current_robot_description.get_arm_chain(designator.arm).get_tool_frame()
        root_link = "map"

        giskard.avoid_all_collisions()
        if designator.allow_gripper_collision:
            giskard.allow_gripper_collision(designator.arm)

        giskard.achieve_cartesian_waypoints_goal(waypoints=waypoints,
                                                 tip_link=tip_link, root_link=root_link,
                                                 enforce_final_orientation=True if designator.movement_type == WaypointsMovementType.ENFORCE_ORIENTATION_FINAL_POINT else False)


class DefaultManager(ProcessModuleManager):

    def __init__(self):
        super().__init__("default")

    def navigate(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultNavigation(self._navigate_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultNavigationReal(self._navigate_lock)

    def looking(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveHead(self._looking_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveHeadReal(self._looking_lock)

    def detecting(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultDetecting(self._detecting_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultDetectingReal(self._detecting_lock)

    def move_tcp(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveTCP(self._move_tcp_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveTCPReal(self._move_tcp_lock)

    def move_arm_joints(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveArmJoints(self._move_arm_joints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveArmJointsReal(self._move_arm_joints_lock)

    def world_state_detecting(self):
        if (ProcessModuleManager.execution_type == ExecutionType.SIMULATED or
                ProcessModuleManager.execution_type == ExecutionType.REAL):
            return DefaultWorldStateDetecting(self._world_state_detecting_lock)

    def move_joints(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveJoints(self._move_joints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveJointsReal(self._move_joints_lock)

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveGripper(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveGripperReal(self._move_gripper_lock)

    def open(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultOpen(self._open_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultOpenReal(self._open_lock)

    def close(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultClose(self._close_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultCloseReal(self._close_lock)

    def move_tcp_waypoints(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveTCPWaypoints(self._move_tcp_waypoints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveTCPWaypointsReal(self._move_tcp_waypoints_lock)

# Initialize the default manager and register it with the ProcessModuleManager
DefaultManager()


from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path

from typing_extensions import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from .datastructures.pose import PoseStamped
    from .description import Link, Joint
    from .world_concepts.world_object import Object
    from .datastructures.enums import JointType, MultiverseAPIName, Arms, StaticJointState, Grasp, DetectionTechnique, \
        ContainerManipulationType
    from .datastructures.world_entity import PhysicalBody
    from .validation.goal_validator import MultiJointPositionGoalValidator
    from .designator import ObjectDesignatorDescription
    from .designators.location_designator import Location


class PlanFailure(Exception):
    """Implementation of plan failures."""

    def __init__(self, *args, **kwargs):
        """Create a new plan failure."""
        Exception.__init__(self, *args, **kwargs)


class KnowledgeNotAvailable(PlanFailure):
    """Thrown when a knowledge source can not provide the information for a query."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NotALanguageExpression(PlanFailure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class FailureDiagnosis(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LowLevelFailure(FailureDiagnosis):
    """Failure thrown by low-level modules: robot or projection PMs."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ActionlibActionTimeout(LowLevelFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class HighLevelFailure(FailureDiagnosis):
    """Failure thrown by high-level modules, i.e. plans."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SensorMonitoringCondition(PlanFailure):
    """Thrown when a sensor monitoring condition is met."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DeliveringFailed(HighLevelFailure):
    """Thrown when delivering plan completely gives up."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ManipulationLowLevelFailure(LowLevelFailure):
    """Thrown when a low-level, i.e. hardware related, failure is detected in a manipulation action."""
    robot: Object
    """
    The robot that the manipulation action was performed with.
    """
    arm: List[Arms]
    """
    The arm(s) that the manipulation action was performed with.
    """
    body: PhysicalBody
    """
    The body that the manipulation action was performed on.
    """

    def __init__(self, robot: Object, arms: List[Arms], body: PhysicalBody, *args, **kwargs):
        self.robot = robot
        self.arm = arms
        self.body = body
        super().__init__(*args, **kwargs)


class EnvironmentManipulationGoalNotReached(ManipulationLowLevelFailure):
    """Thrown when door / drawer opening / closing goal is still not reached."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ContainerManipulationError(ManipulationLowLevelFailure, ABC):
    """Thrown when container manipulation fails."""
    container_joint: Joint
    """
    The joint of the container that should be manipulated.
    """
    manipulation_type: ContainerManipulationType
    """
    The type of manipulation that should be performed on the container.
    """

    def __init__(self, robot: Object, arms: List[Arms], body: PhysicalBody, container_joint: Joint,
                 manipulation_type: ContainerManipulationType,
                 *args, **kwargs):
        self.container_joint = container_joint
        self.manipulation_type = manipulation_type
        super().__init__(robot, arms, body,
                         f"manipulation type \"{manipulation_type.name}\" failedContainer {body.name} with joint"
                         f" name {container_joint.name}"
                         f" (current position is {container_joint.position}) and limits {container_joint.limits},"
                         f" using {[arm.name for arm in arms]} arm of {robot.name} robot", *args, **kwargs)


class EnvironmentManipulationImpossible(HighLevelFailure):
    """Thrown when environment manipulation cannot be achieved."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class EnvironmentUnreachable(HighLevelFailure):
    """Thrown when environment manipulation in collision or unreachable."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class FetchingFailed(HighLevelFailure):
    """Thrown when fetching plan completely gives up."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class GripperLowLevelFailure(LowLevelFailure):
    """Thrown when a failure involving the gripper hardware occurs."""
    robot: Object
    """
    The robot that the gripper belongs to.
    """
    arm: Arms
    """
    The arm that the gripper belongs to.
    """

    def __init__(self, robot: Object, arm: Arms, *args, **kwargs):
        self.robot = robot
        self.arm = arm
        super().__init__(*args, **kwargs)


class GripperIsNotOpen(GripperLowLevelFailure):
    """Thrown when the gripper is not open when it should be open."""

    def __init__(self, robot: Object, arm: Arms, *args, **kwargs):
        super().__init__(robot, arm,
                         f"The gripper of arm {arm.name} of robot {robot.name} should be open but is not",
                         *args, **kwargs)


class GripperClosedCompletely(GripperLowLevelFailure):
    """Thrown when the gripper closed completely, despite not being expected to do so (e.g. because it should have
    grasped something)."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class GripperGoalNotReached(GripperLowLevelFailure):
    """Thrown when the gripper does not reach its goal."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class GripperOccupied(GripperLowLevelFailure):
    """Thrown when the gripper is occupied by some object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LookingHighLevelFailure(HighLevelFailure):
    """High-level failure produced when looking for an object, i.e. it is not a hardware issue but one relating to
    the looking task, its parameters, and how they relate to the environment."""
    robot: Object
    """
    The robot that performed the look at action.
    """
    target: PoseStamped
    """
    The target pose that the robot was supposed to look at.
    """

    def __init__(self, robot: Object, target: PoseStamped, *args, **kwargs):
        self.robot = robot
        self.target = target
        super().__init__(*args, **kwargs)


class LookAtGoalNotReached(LookingHighLevelFailure):
    """Thrown when the look at goal is not reached."""

    def __init__(self, robot: Object, target: PoseStamped, *args, **kwargs):
        super().__init__(robot, target, f"Look at action failed for {robot.name} and target "
                                        f"{target.position.to_list()}{target.orientation.to_list()}", *args, **kwargs)


class ManipulationGoalInCollision(HighLevelFailure):
    """Thrown when executing a manipulation action results in a collision."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ManipulationGoalNotReached(ManipulationLowLevelFailure):
    """Thrown when after executing the action, goal is still not reached."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class RobotInCollision(PlanFailure):
    """Thrown when the robot is in collision with the environment."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class IKError(PlanFailure):
    """Thrown when no inverse kinematics solution could be found"""
    pose: PoseStamped
    """
    The pose for which no IK solution could be found.
    """
    base_frame: str
    """
    The base frame in which the pose was given.
    """
    tip_frame: str
    """
    The robot tip frame that should reach the pose.
    """

    def __init__(self, pose, base_frame, tip_frame):
        self.pose = pose
        self.base_frame = base_frame
        self.tip_frame = tip_frame
        self.message = "Position {} in frame '{}' is not reachable for end effector: '{}'".format(pose, base_frame,
                                                                                                  tip_frame)
        super(IKError, self).__init__(self.message)


class ManipulationPoseUnreachable(ManipulationLowLevelFailure):
    """Thrown when no IK solution can be found."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NavigationHighLevelFailure(HighLevelFailure):
    """High-level failure produced while navigating the robot, i.e. it is not a hardware issue but one relating to
    the navigation task, its parameters, and how they relate to the environment."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NavigationGoalInCollision(NavigationHighLevelFailure):
    """Navigation goal cannot be reached because the goal itself is already occupied by some other object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NavigationLowLevelFailure(LowLevelFailure):
    """Low-level failure produced while navigating the robot, i.e. some kind of hardware issue."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NavigationGoalNotReached(NavigationLowLevelFailure):
    """Thrown when the base moved as a result of the navigation action but the goal was not reached."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NavigationPoseUnreachable(NavigationLowLevelFailure):
    """Thrown when the goal pose for navigation is computed to be unreachable."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectNotVisible(HighLevelFailure):
    """Thrown when the robot cannot see an object of a given description in its surroundings."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectNotFound(HighLevelFailure):
    """Thrown when the robot cannot find an object of a given description in its surroundings."""

    def __init__(self, object_name: str):
        super().__init__(f"Object {object_name} not found")


class LinkNotFound(HighLevelFailure):
    """Thrown when the robot cannot find a link of a given description"""
    link_name: str
    """
    The name of the link that couldn't be found
    """
    of_object: Optional[str]
    """
    The name of the object that the link should belong to
    """

    def __init__(self, link_name: str, of_object: Optional[str] = None):
        self.link_name = link_name
        self.of_object = of_object if of_object else None
        of_object_str = f" of object {of_object}" if of_object else ""
        super().__init__(f"Link {self.link_name} not found" + of_object_str)


class ObjectUndeliverable(HighLevelFailure):
    """Thrown when no base positioning can assure a reachable pose to place the object from."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectPlacingError(HighLevelFailure):
    """Thrown when the placing of the object fails."""
    obj: Object
    """
    The object that should be placed.
    """
    placing_pose: PoseStamped
    """
    The target pose at which the object should be placed.
    """
    robot: Object
    """
    The robot that placed the object.
    """
    arm: Arms
    """
    The robot arm used to place the object.
    """

    def __init__(self, obj: Object, placing_pose: PoseStamped, robot: Object, arm: Arms,
                 *args, **kwargs):
        self.obj = obj
        self.placing_pose = placing_pose
        self.robot = robot
        self.arm = arm
        super().__init__(*args, **kwargs)


class ObjectStillInContact(ObjectPlacingError):
    """Thrown when the object is still in contact with the robot after placing."""
    contact_links: List[Link]
    """
    The links of the robot that are still in contact with the object.
    """

    def __init__(self, obj: Object, contact_links: List[Link], placing_pose: PoseStamped, robot: Object, arm: Arms,
                 *args, **kwargs):
        self.contact_links = contact_links
        contact_link_names = [link.name for link in contact_links]
        super().__init__(obj, placing_pose, robot, arm,
                         f"Object {obj.name} is still in contact with {robot.name}, the contact links are"
                         f"{contact_link_names}, after placing at"
                         f" target pose {placing_pose.position.to_list()}{placing_pose.orientation.to_list()} using"
                         f" {arm.name} arm", *args, **kwargs)


class ObjectNotPlacedAtTargetLocation(ObjectPlacingError):
    """Thrown when the object was not placed at the target location."""

    def __init__(self, obj: Object, placing_pose: PoseStamped, robot: Object, arm: Arms, *args, **kwargs):
        super().__init__(obj, placing_pose, robot, arm,
                         "Object {obj.name} was not placed at target pose {placing_pose.position.to_list()}"
                         f"{placing_pose.orientation.to_list()} using {arm.name} arm of {robot.name}", *args, **kwargs)


class ObjectUnfetchable(HighLevelFailure):
    """Thrown when no base positioning can assure a reachable pose to grasp the object from."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectUnreachable(HighLevelFailure):
    """Thrown when no IK found for particular base pose."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class PerceptionLowLevelFailure(LowLevelFailure):
    """Low-level failure produced while perceiving, i.e. some kind of hardware issue."""
    object_description: ObjectDesignatorDescription
    """
    The object description that was used to search for the object.
    """
    technique: DetectionTechnique
    """
    The detection technique that was used to search for the object.
    """
    region: Optional[Location] = None
    """
    The suggested region in which the object was searched.
    """

    def __init__(self, object_description: ObjectDesignatorDescription, technique: DetectionTechnique,
                 region: Optional[Location] = None, *args, **kwargs):
        self.object_description = object_description
        self.technique = technique
        self.region = region
        super().__init__(*args, **kwargs)


class PerceptionObjectNotFound(PerceptionLowLevelFailure):
    """Thrown when an attempt to find an object by perception fails -- and this can still be interpreted as the robot
    not looking in the right direction, as opposed to the object being absent."""

    def __init__(self, obj_desc: ObjectDesignatorDescription, technique: DetectionTechnique, region: Location,
                 *args, **kwargs):
        super().__init__(obj_desc, technique, region,
                         f"object described by {obj_desc} not found using {technique.name} technique in region"
                         f" {region}", *args, **kwargs)


class PerceptionObjectNotInWorld(PerceptionLowLevelFailure):
    """Thrown when an attempt to find an object by perception fails -- and this is because the object can be assumed
    absent or perhaps is known absent because of the setup of a simulation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SearchingFailed(HighLevelFailure):
    """Thrown when searching plan completely gives up."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TorsoLowLevelFailure(LowLevelFailure):
    """Low-level failure produced while moving the torso, i.e. some kind of hardware issue."""
    goal_validator: Optional[MultiJointPositionGoalValidator] = None
    """
    The goal validator that was used to check if the goal was reached.
    """

    def __init__(self, goal_validator: Optional[MultiJointPositionGoalValidator] = None, *args, **kwargs):
        self.goal_validator = goal_validator
        if goal_validator:
            super().__init__(goal_validator.goal_not_achieved_message, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)


class TorsoGoalNotReached(TorsoLowLevelFailure):
    """Thrown when the torso moved as a result of a torso action but the goal was not reached."""


class TorsoGoalUnreachable(TorsoLowLevelFailure):
    """Thrown when the goal for the torso is computed to be unreachable."""


class Task(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Grasping(Task):
    """"""
    obj: Object
    """
    The object to be grasped.
    """
    robot: Object
    """
    The robot that should grasp the object.
    """
    arm: Arms
    """
    The arm used to grasp the object.
    """
    grasp: Optional[Grasp]
    """
    The grasp type used to grasp the object.
    """

    def __init__(self, obj: Object, robot: Object, arm: Arms, grasp: Optional[Grasp] = None, *args, **kwargs):
        self.obj = obj
        self.robot = robot
        self.arm = arm
        self.grasp = grasp
        super().__init__(*args, **kwargs)


class ObjectNotGraspedError(Grasping):
    def __init__(self, obj: Object, robot: Object, arm: Arms, grasp = None, *args, **kwargs):
        grasp_str = f"using {grasp} grasp" if grasp else ""
        super().__init__(obj, robot, arm, grasp, f"object {obj.name} was not grasped by {arm.name} arm" + grasp_str,
                         *args, **kwargs)


class Looking(Task):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectPoseMissestimation(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SuccessfulCompletion(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LocomotorFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ArmFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectLost(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SensorFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class IllPosedGoalFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CapabilityAbsenceFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ReachabilityFailure(PlanFailure):
    """"""
    obj: Object
    """
    The object that should be reachable.
    """
    robot: Object
    """
    The robot that should reach the object.
    """
    arm: Arms
    """
    The arm that should reach the object.
    """
    grasp: Optional[Grasp] = None
    """
    The grasp/gripper orientation that should be used to reach the object.
    """

    def __init__(self, obj: Object, robot: Object, arm: Arms, grasp: Grasp, *args, **kwargs):
        self.obj = obj
        self.robot = robot
        self.arm = arm
        self.grasp = grasp
        super().__init__(*args, **kwargs)


class ObjectNotInGraspingArea(ReachabilityFailure):
    def __init__(self, obj: Object, robot: Object, arm: Arms, grasp, *args, **kwargs):
        super().__init__(obj, robot, arm, grasp,
                         f"object {obj.name} is not in the grasping area of robot {robot.name} using {arm.name} arm and"
                         f" {grasp} grasp", *args, **kwargs)


class TorsoFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ConfigurationNotReached(PlanFailure):
    """"""
    goal_validator: MultiJointPositionGoalValidator
    """
    The goal validator that was used to check if the goal was reached.
    """
    configuration_type: StaticJointState
    """
    The configuration type that should be reached.
    """

    def __init__(self, goal_validator: MultiJointPositionGoalValidator, configuration_type: StaticJointState,
                 *args, **kwargs):
        self.goal_validator = goal_validator
        self.configuration_type = configuration_type
        super().__init__(f"configuration_type: {configuration_type.name},"
                         f" {goal_validator.goal_not_achieved_message}",
                         *args, **kwargs)


class Timeout(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class EndEffectorFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ObjectUnavailable(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SustainedFailure(PlanFailure):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ReasoningError(PlanFailure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CollisionError(PlanFailure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NavigationGoalNotReachedError(PlanFailure):
    """
    Thrown when the navigation goal is not reached.
    """
    current_pose: PoseStamped
    """
    The current pose of the robot.
    """
    goal_pose: PoseStamped
    """
    The goal pose of the robot.
    """

    def __init__(self, current_pose: PoseStamped, goal_pose: PoseStamped, *args, **kwargs):
        self.current_pose = current_pose
        self.goal_pose = goal_pose
        super().__init__(f"Navigation goal not reached. Current pose: {current_pose}, goal pose: {goal_pose}",
                         *args, **kwargs)


class ToolPoseNotReachedError(PlanFailure):
    """
    Thrown when the tool pose is not reached.
    """
    current_pose: PoseStamped
    """
    The current pose of the tool.
    """
    goal_pose: PoseStamped
    """
    The goal pose of the tool.
    """

    def __init__(self, current_pose: PoseStamped, goal_pose: PoseStamped, *args, **kwargs):
        self.current_pose = current_pose
        self.goal_pose = goal_pose
        super().__init__(f"Tool pose not reached. Current pose: {current_pose}, goal pose: {goal_pose}",
                         *args, **kwargs)


"""
The following exceptions are used in the PyCRAM framework to handle errors related to the world and the objects in it.
They are usually related to a bug in the code or a misuse of the framework (e.g. logical errors in the code).
"""


class MultiverseFailedAPIResponse(Exception):
    """
    Exception raised when a Multiverse API call fails.
    """
    api_response: List[str]
    """
    The response of the API call that failed.
    """
    api_name: MultiverseAPIName
    """
    The name of the API that failed.
    """
    def __init__(self, api_response: List[str], api_name: MultiverseAPIName, *args, **kwargs):
        self.api_response = api_response
        self.api_name = api_name
        super().__init__(f"{api_name} api request with arguments {args} and keyword arguments {kwargs}"
                         f" failed with response {api_response}")


class ProspectionObjectNotFound(KeyError):
    """
    Exception raised when an object was not found in the prospection world.
    """
    obj: Object
    """
    The object that was not found in the prospection world.
    """
    def __init__(self, obj: Object):
        self.obj = obj
        super().__init__(f"The given object {obj.name} is not in the prospection world.")


class ObjectAlreadyExists(Exception):
    """
    Exception raised when an object with the same name already exists in the world.
    """
    obj: Object
    """
    The object that already exists in the world.
    """
    def __init__(self, obj: Object):
        self.obj = obj
        super().__init__(f"An object with the name {obj.name} already exists in the world.")


class ObjectDescriptionNotFound(KeyError):
    """
    Exception raised when the description of an object was not found.
    """
    object_name: str
    """
    The name of the object whose description was not found.
    """
    path: str
    """
    The path of the object whose description was not found.
    """
    extension: str
    """
    The description extension of the object whose description was not found.
    """
    def __init__(self, object_name: str, path: str, extension: str):
        self.object_name = object_name
        self.path = path
        self.extension = extension
        super().__init__(f"{object_name} with path {path} and extension {extension} is not in supported extensions, and"
                         f" the description data was not found on the ROS parameter server")


class WorldMismatchErrorBetweenAttachedObjects(Exception):
    """
    Exception raised when two objects that are attached to each other have a mismatch in the world they belong to.
    """
    obj_1: Object
    """
    The first object that has a mismatch in the world.
    """
    obj_2: Object
    """
    The second object that has a mismatch in the world.
    """
    def __init__(self, obj_1: 'Object', obj_2: 'Object'):
        self.obj_1 = obj_1
        self.obj_2 = obj_2
        super().__init__(f"World mismatch between the attached objects {obj_1.name} and {obj_2.name},"
                         f"obj_1.world: {obj_1.world}, obj_2.world: {obj_2.world}")


class ObjectFrameNotFoundError(KeyError):
    """
    Exception raised when a tf frame of an object is not found.
    """
    frame_name: str
    """
    The name of the frame that was not found.
    """
    def __init__(self, frame_name: str):
        self.frame_name = frame_name
        super().__init__(f"Frame {frame_name} does not belong to any of the objects in the world.")


class MultiplePossibleTipLinks(Exception):
    """
    Exception raised when multiple tip links are found for an object.
    """
    object_name: str
    """
    The name of the object that has multiple tip links.
    """
    start_link: str
    """
    The start link of the object that has multiple tip links.
    """
    tip_links: List[str]
    """
    The list of tip links that are found for the object.
    """
    def __init__(self, object_name: str, start_link: str, tip_links: List[str]):
        self.object_name = object_name
        self.start_link = start_link
        self.tip_links = tip_links
        super().__init__(f"Multiple possible tip links found for object {object_name} with start link {start_link}:"
                         f" {tip_links}")


class UnsupportedFileExtension(Exception):
    """
    Exception raised when an object mesh/description has an unsupported file extension.
    """
    object_name: str
    """
    The name of the object that has an unsupported file extension.
    """
    path: str
    """
    The path of the object description/mesh that has an unsupported file extension.
    """
    extension: str
    """
    The unsupported file extension of the object description/mesh.
    """
    def __init__(self, object_name: str, path: str):
        self.object_name = object_name
        self.path = path
        self.extension = Path(path).suffix
        super().__init__(f"Unsupported file extension for object {object_name} with path {path}"
                         f"and extension {self.extension}")


class ObjectDescriptionUndefined(Exception):
    """
    Exception raised when the given object description type is not defined or couldn't be resolved to a known type.
    """
    object_name: str
    """
    The name of the object that has an undefined description.
    """
    def __init__(self, object_name: str):
        self.object_name = object_name
        super().__init__(f"Object description for object {object_name} is not defined, either a path or a description"
                         f"object should be provided.")


class UnsupportedJointType(Exception):
    """
    Exception raised when an unsupported joint type is used.
    """
    joint_type: JointType
    """
    The unsupported joint type that was used.
    """
    def __init__(self, joint_type: JointType):
        self.joint_type = joint_type
        super().__init__(f"Unsupported joint type: {joint_type}")


class LinkHasNoGeometry(Exception):
    """
    Exception raised when a link has no geometry (i.e. no visual or collision elements).
    """
    link_name: str
    """
    The name of the link that has no geometry.
    """
    def __init__(self, link_name: str):
        self.link_name = link_name
        super().__init__(f"Link {link_name} has no geometry.")


class LinkGeometryHasNoMesh(Exception):
    """
    Exception raised when a link geometry has no mesh or is not of type mesh.
    """
    link_name: str
    """
    The name of the link that has no mesh.
    """
    geometry_type: str
    """
    The type of the link geometry.
    """
    def __init__(self, link_name: str, geometry_type: str):
        self.link_name = link_name
        self.geometry_type = geometry_type
        super().__init__(f"Link {link_name} geometry with type {geometry_type} has no mesh.")



from . import node
import rclpy

def create_subscriber(topic, msg_type, callback, queue_size=10) -> rclpy.subscription.Subscription:
    node.create_subscription(msg_type, topic, callback, queue_size)


from typing_extensions import Any

from .default_process_modules import *
from ..datastructures.enums import GripperState, ExecutionType
from ..datastructures.world import World
from ..designators.motion_designator import MoveGripperMotion
from ..process_module import ProcessModule, ProcessModuleManager
from ..robot_descriptions.ur5e_controlled_description import data as ur5e_data
from ..ros import create_publisher

try:
    from std_msgs.msg import Float64
except ImportError:
    loginfo ("Float64 message type not found, make sure to install std_msgs package")

class RobotiqMoveGripperReal(DefaultMoveGripperReal):
    """
    Opens or closes the gripper of the real Robotiq gripper, uses a topic for this instead of giskard
    """

    def _execute(self, designator: MoveGripperMotion) -> Any:
        value = ur5e_data.gripper_open_cmd_value if designator.motion == GripperState.OPEN \
            else ur5e_data.gripper_close_cmd_value
        publisher = create_publisher(ur5e_data.gripper_cmd_topic, Float64, queue_size=10, latch=True)
        World.current_world.step(func=lambda: publisher.publish(Float64(value)), step_seconds=0.1)
        return True


class RobotiqManager(DefaultManager):

    def __init__(self):
        super().__init__()
        self.robot_name = ur5e_data.gripper_name

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveGripper(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return RobotiqMoveGripperReal(self._move_gripper_lock)

RobotiqManager()



class DeprecatedRobotDescription:
    def raise_error(self):
        raise DeprecationWarning("Robot description moved, please use RobotDescription.current_robot_description from"
                                 " pycram.robot_description")

    @property
    def name(self):
        self.raise_error()

    @property
    def chains(self):
        self.raise_error()

    @property
    def joints(self):
        self.raise_error()

    @property
    def torso_joint(self):
        self.raise_error()

    @property
    def torso_link(self):
        self.raise_error()


robot_description = DeprecatedRobotDescription()



from dataclasses import dataclass

import sqlalchemy.orm
import sqlalchemy.sql
from sqlalchemy import select, Select
from typing_extensions import List, Type
from ....costmaps import Rectangle, OccupancyCostmap
from ....designator import LocationDesignatorDescription
from ....designators.location_designator import CostmapLocation
from ....orm.views import PickUpWithContextView
from ....datastructures.pose import PoseStamped


@dataclass
class Location(LocationDesignatorDescription.Location):
    """
    A location that is described by a pose, a reachable arm, a torso height and a grasp.
    """
    pose: PoseStamped
    reachable_arm: str
    torso_height: float
    grasp: str


class AbstractCostmapLocation(CostmapLocation):
    """
    Abstract Class for JPT and Database costmaps.
    """

    def __init__(self, target, reachable_for=None, reachable_arm=None):
        """
        Create a new AbstractCostmapLocation instance.
        :param target: The target object
        :param reachable_for:
        :param reachable_arm:
        """
        super().__init__(target, reachable_for, None, reachable_arm, None)

    def create_occupancy_rectangles(self) -> List[Rectangle]:
        """
        :return: A list of rectangles that represent the occupied space of the target object.
        """
        # create Occupancy costmap for the target object
        ocm = OccupancyCostmap(distance_to_obstacle=0.3, from_ros=False, size=200, resolution=0.02,
                               origin=self.target.pose)
        return ocm.partitioning_rectangles()


class DatabaseCostmapLocation(AbstractCostmapLocation):
    """
    Class that represents costmap locations from a given Database.
    The database has to have a schema that is compatible with the pycram.orm package.
    """

    def __init__(self, target, session: sqlalchemy.orm.Session = None, reachable_for=None, reachable_arm=None):
        """
        Create a Database Costmap

        :param target: The target object
        :param session: A session that can be used to execute queries
        :param reachable_for: The robot to grab the object with
        :param reachable_arm: The arm to use

        """
        super().__init__(target, reachable_for, reachable_arm)
        self.session = session

    @staticmethod
    def select_statement(view: Type[PickUpWithContextView]) -> Select:
        return (select(view.arm, view.grasp, view.torso_height, view.relative_x, view.relative_y, view.quaternion_x,
                       view.quaternion_y, view.quaternion_z, view.quaternion_w).distinct())

    def create_query_from_occupancy_costmap(self) -> Select:
        """
        Create a query that queries all relative robot positions from an object that are not occluded using an
        OccupancyCostmap.
        """

        view = PickUpWithContextView

        # get query
        query = self.select_statement(view)

        # constraint query to correct object type and successful task status
        query = query.where(view.obj_type == self.target.obj_type).where(view.status == "SUCCEEDED")

        filters = []

        # for every rectangle
        for rectangle in self.create_occupancy_rectangles():
            # add sql filter
            filters.append(sqlalchemy.and_(view.relative_x >= rectangle.x_lower,
                                           view.relative_x < rectangle.x_upper,
                                           view.relative_y >= rectangle.y_lower,
                                           view.relative_y < rectangle.y_upper))

        return query.where(sqlalchemy.or_(*filters))

    def sample_to_location(self, sample: sqlalchemy.engine.row.Row) -> Location:
        """
        Convert a database row to a costmap location.

        :param sample: The database row.
        :return: The costmap location
        """

        target_x, target_y, target_z = self.target.pose.position.to_list()
        position = [target_x + sample[3], target_y + sample[4], 0]
        orientation = [sample[5], sample[6], sample[7], sample[8]]

        result = Location(PoseStamped(position, orientation), sample.arm, sample.torso_height, sample.grasp)
        return result

    def __iter__(self) -> Location:
        statement = self.create_query_from_occupancy_costmap().limit(200)
        samples = self.session.execute(statement).all()
        if samples:
            for sample in samples:
                yield self.sample_to_location(sample)
        else:
            raise ValueError("No samples found")



from __future__ import annotations

import copy
import os
import pathlib
import xml.etree.ElementTree as ET

import numpy as np
from ..tf_transformations import quaternion_from_euler, euler_from_quaternion
from typing_extensions import Union, List, Optional, Dict, Tuple, Type, Self
from urdf_parser_py import urdf
from urdf_parser_py.urdf import (URDF, Collision, Box as URDF_Box, Cylinder as URDF_Cylinder,
                                 Sphere as URDF_Sphere, Mesh as URDF_Mesh)

from ..ros import get_ros_package_path
from ..datastructures.dataclasses import Color, VisualShape, BoxVisualShape, CylinderVisualShape, \
    SphereVisualShape, MeshVisualShape
from ..datastructures.enums import JointType
from ..datastructures.pose import PoseStamped, Point
from ..description import JointDescription as AbstractJointDescription, \
    LinkDescription as AbstractLinkDescription, ObjectDescription as AbstractObjectDescription
from ..failures import MultiplePossibleTipLinks
from ..ros import  logerr
from ..ros import  create_ros_pack, ResourceNotFound, get_parameter
from ..utils import suppress_stdout_stderr


class LinkDescription(AbstractLinkDescription):
    """
    A class that represents a link description of an object.
    """

    def __init__(self, urdf_description: urdf.Link):
        super().__init__(urdf_description)
        # match the name of the link to the class in the ontology that this may be


    @property
    def geometry(self) -> List[VisualShape]:
        """
        :return: The geometry type of the URDF collision element of this link.
        """
        if self.collision is None:
            return []
        if isinstance(self.collision, List):
            return [self._get_visual_shape(coll.geometry) for coll in self.collision]
        urdf_geometry = self.collision.geometry
        return [self._get_visual_shape(urdf_geometry)]

    @property
    def visual_geometry(self) -> List[VisualShape]:
        """
        :return: The geometry type of the URDF visual element of this link.
        """
        visuals = self.parsed_description.visuals
        return [self._get_visual_shape(vis.geometry) for vis in visuals]

    @staticmethod
    def _get_visual_shape(urdf_geometry) -> Union[VisualShape, None]:
        """
        :param urdf_geometry: The URDFGeometry for which the visual shape is returned.
        :return: the VisualShape of the given URDF geometry.
        """
        if isinstance(urdf_geometry, URDF_Box):
            half_extents = np.array(urdf_geometry.size) / 2
            return BoxVisualShape(Color(), [0, 0, 0], half_extents.tolist())
        if isinstance(urdf_geometry, URDF_Cylinder):
            return CylinderVisualShape(Color(), [0, 0, 0], urdf_geometry.radius, urdf_geometry.length)
        if isinstance(urdf_geometry, URDF_Sphere):
            return SphereVisualShape(Color(), [0, 0, 0], urdf_geometry.radius)
        if isinstance(urdf_geometry, URDF_Mesh):
            return MeshVisualShape(Color(), [0, 0, 0], urdf_geometry.scale, urdf_geometry.filename)
        return None

    @property
    def origin(self) -> Optional[PoseStamped]:
        """
        The origin of this link

        :return: The origin of this link as a PoseStamped object.
        """
        if self.collision is None:
            return None
        coll = self.collision[0] if isinstance(self.collision, List) else self.collision
        if coll.origin is None:
            return None
        return PoseStamped.from_list(coll.origin.xyz,
                           quaternion_from_euler(*coll.origin.rpy))

    @property
    def name(self) -> str:
        return self.parsed_description.name

    @property
    def collision(self) -> Union[Collision, List[Collision], None]:
        if self.parsed_description.collisions:
            if len(self.parsed_description.collisions) == 1:
                return self.parsed_description.collisions[0]
            else:
                return self.parsed_description.collisions

    @property
    def all_collisions(self) -> List[Collision]:
        return self.parsed_description.collisions


class JointDescription(AbstractJointDescription):
    urdf_type_map = {'unknown': JointType.UNKNOWN,
                     'revolute': JointType.REVOLUTE,
                     'continuous': JointType.CONTINUOUS,
                     'prismatic': JointType.PRISMATIC,
                     'floating': JointType.FLOATING,
                     'planar': JointType.PLANAR,
                     'fixed': JointType.FIXED}

    pycram_type_map = {pycram_type: urdf_type for urdf_type, pycram_type in urdf_type_map.items()}

    def __init__(self, urdf_description: urdf.Joint, is_virtual: Optional[bool] = False):
        super().__init__(urdf_description, is_virtual=is_virtual)

    @property
    def origin(self) -> PoseStamped:
        return PoseStamped.from_list(self.parsed_description.origin.xyz,
                                    quaternion_from_euler(*self.parsed_description.origin.rpy))

    @property
    def name(self) -> str:
        return self.parsed_description.name

    @property
    def has_limits(self) -> bool:
        return bool(self.parsed_description.limit) and not self.type == JointType.CONTINUOUS

    @property
    def type(self) -> JointType:
        """
        :return: The type of this joint.
        """
        return self.urdf_type_map[self.parsed_description.type]

    @property
    def axis(self) -> Point:
        """
        :return: The axis of this joint, for example the rotation axis for a revolute joint.
        """
        return Point(**dict(zip(["x", "y", "z"],self.parsed_description.axis)))

    @property
    def lower_limit(self) -> Union[float, None]:
        """
        :return: The lower limit of this joint, or None if the joint has no limits.
        """
        if self.has_limits:
            return self.parsed_description.limit.lower
        else:
            return None

    @property
    def upper_limit(self) -> Union[float, None]:
        """
        :return: The upper limit of this joint, or None if the joint has no limits.
        """
        if self.has_limits:
            return self.parsed_description.limit.upper
        else:
            return None

    @property
    def parent(self) -> str:
        """
        :return: The name of the parent link of this joint.
        """
        return self.parsed_description.parent

    @property
    def child(self) -> str:
        """
        :return: The name of the child link of this joint.
        """
        return self.parsed_description.child

    @property
    def damping(self) -> float:
        """
        :return: The damping of this joint.
        """
        return self.parsed_description.dynamics.damping

    @property
    def friction(self) -> float:
        """
        :return: The friction of this joint.
        """
        return self.parsed_description.dynamics.friction


class ObjectDescription(AbstractObjectDescription):
    """
    A class that represents an object description of an object.
    """

    class Link(AbstractObjectDescription.Link, LinkDescription):
        ...

    class RootLink(AbstractObjectDescription.RootLink, Link):
        ...

    class Joint(AbstractObjectDescription.Joint, JointDescription):
        ...

    @property
    def child_map(self) -> Dict[str, List[Tuple[str, str]]]:
        """
        :return: A dictionary mapping the name of a link to its children which are represented as a tuple of the child
            joint name and the link name.
        """
        return self.parsed_description.child_map

    @property
    def parent_map(self) -> Dict[str, Tuple[str, str]]:
        """
        :return: A dictionary mapping the name of a link to its parent joint and link as a tuple.
        """
        return self.parsed_description.parent_map

    @property
    def link_map(self) -> Dict[str, LinkDescription]:
        """
        :return: A dictionary mapping the name of a link to its description.
        """
        if self._link_map is None:
            self._init_links_map()
        return self._link_map

    @property
    def joint_map(self) -> Dict[str, JointDescription]:
        """
        :return: A dictionary mapping the name of a joint to its description.
        """
        if self._joint_map is None:
            self._init_joints_map()
        return self._joint_map

    def _init_joints_map(self) -> None:
        self._joint_map = {joint.name: joint for joint in self.joints}

    def _init_links_map(self) -> None:
        self._link_map = {link.name: link for link in self.links}

    def add_joint(self, name: str, child: str, joint_type: JointType,
                  axis: Optional[Point] = None, parent: Optional[str] = None, origin: Optional[PoseStamped] = None,
                  lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                  is_virtual: Optional[bool] = False) -> None:
        """
        Add a joint to the object description, could be a virtual joint as well.
        For documentation of the parameters, see :meth:`pycram.description.ObjectDescription.add_joint`.
        """
        if lower_limit is not None or upper_limit is not None:
            limit = urdf.JointLimit(lower=lower_limit, upper=upper_limit)
        else:
            limit = None
        if origin is not None:
            origin = urdf.Pose(origin.position.to_list(), euler_from_quaternion(origin.orientation.to_list()))
        if axis is not None:
            axis = [axis.x, axis.y, axis.z]
        if parent is None:
            parent = self.get_root()
        if is_virtual:
            self.virtual_joint_names.append(name)
        else:
            joint = urdf.Joint(name,
                               parent,
                               child,
                               JointDescription.pycram_type_map[joint_type],
                               axis, origin, limit)
            self.add_joint_from_parsed_description(joint)

    def add_joint_from_parsed_description(self, description: urdf.Joint) -> None:
        """
        Add a joint to the object description from a parsed URDF joint description.

        :param description: The parsed URDF joint description.
        """
        self.parsed_description.add_joint(description)
        if self._joints is not None:
            self._joints.append(JointDescription(description))
            self._joint_map[description.name] = self._joints[-1]

    def add_link_from_parsed_description(self, description: urdf.Link) -> None:
        """
        Add a link to the object description from a parsed URDF link description.

        :param description: The parsed URDF link description.
        """
        self.parsed_description.add_link(description)
        if self._links is not None and len(self._links) > 0:
            self._links.append(LinkDescription(description))
            self._link_map[description.name] = self._links[-1]

    def merge_description(self, other: ObjectDescription, parent_link: Optional[str] = None,
                          child_link: Optional[str] = None,
                          joint_type: JointType = JointType.FIXED,
                          axis: Optional[Point] = None,
                          lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                          child_pose_wrt_parent: Optional[PoseStamped] = None,
                          in_place: bool = False,
                          new_description_file: Optional[str] = None) -> Union[ObjectDescription, Self]:
        other_description = other.parsed_description
        if child_pose_wrt_parent is None:
            child_pose_wrt_parent = PoseStamped()

        original_child_link = child_link if child_link is not None else other.get_root()
        child_link = f"{other_description.name}_" + original_child_link
        parent_link = parent_link if parent_link is not None else self.get_root()

        description = self if in_place else copy.deepcopy(self)

        # Add the links of the other description to this description
        for link in other_description.links:
            # Change the name of the link to avoid name conflicts
            link = copy.deepcopy(link)
            link.name = f'{other_description.name}_{link.name}'
            description.add_link_from_parsed_description(link)

        # Add the joints of the other description to this description
        for joint in other_description.joints:
            # Change the name of the joint to avoid name conflicts
            joint = copy.deepcopy(joint)
            joint.name = f'{other_description.name}_{joint.name}'
            description.add_joint_from_parsed_description(joint)

        # Add the joint between the parent and child link
        description.add_joint(f'{parent_link}_{original_child_link}_joint', child_link, joint_type,
                              axis=axis, origin=child_pose_wrt_parent, parent=parent_link,
                              lower_limit=lower_limit, upper_limit=upper_limit)

        if new_description_file is not None:
            description.xml_path = os.path.join(pathlib.Path(description.xml_path).parent,
                                                (new_description_file + description.get_file_extension()))
        description.write_description_to_file(description.parsed_description.to_xml_string(), description.xml_path)
        return description

    def _init_description(self) -> None:
        self._init_links()
        self._init_joints()

    def load_description(self, path) -> URDF:
        with open(path, 'r') as file:
            # Since parsing URDF causes a lot of warning messages which can't be deactivated, we suppress them
            with suppress_stdout_stderr():
                return URDF.from_xml_string(file.read())

    def generate_from_mesh_file(self, path: str, name: str, save_path: str, color: Optional[Color] = Color(),
                                scale: Optional[float] = None) -> None:
        """
        Generate a URDf file with the given .obj or .stl file as mesh. In addition, use the given rgba_color to create a
         material tag in the URDF. The URDF file will be saved to the given save_path.

        :param path: The path to the mesh file.
        :param name: The name of the object.
        :param save_path: The path to save the URDF file to.
        :param color: The color of the object.
        :param scale: The scale of the mesh.
        """
        urdf_template = '<?xml version="0.0" ?> \n \
                        <robot name="~a_object"> \n \
                        <material name="color">\n \
                                <color rgba="~c"/>\n \
                        </material>\n \
                         <link name="~a_main"> \n \
                            <visual> \n \
                                <geometry>\n \
                                    <mesh filename="~b" scale="~d"/> \n \
                                </geometry>\n \
                                <material name="color"/>\n \
                          </visual> \n \
                        <collision> \n \
                            <geometry>\n \
                                <mesh filename="~b" scale="~d"/>\n \
                            </geometry>\n \
                            <material name="color"/>\n \
                        </collision>\n \
                        </link> \n \
                        </robot>'
        urdf_template = self.fix_missing_inertial(urdf_template)
        rgb = " ".join(list(map(str, color.get_rgba())))
        scale = scale if scale is not None else 1
        s = " ".join([str(scale)] * 3)
        pathlib_obj = pathlib.Path(path)
        path = str(pathlib_obj.resolve())
        content = urdf_template.replace("~a", name).replace("~b", path).replace("~c", rgb).replace("~d", s)
        self.write_description_to_file(content, save_path)

    def generate_from_description_file(self, path: str, save_path: str, make_mesh_paths_absolute: bool = True) -> None:
        """
        Generates a loadable URDF file from the given URDF file. The method will fix the attributes of some links,
        replace relative paths with absolute paths and fix the missing inertial tags. The generated URDF file will be
        saved to the given save_path.

        :param path: Path to the URDF file which should be processed.
        :param save_path: Path to where to save the processed URDF file.
        :param make_mesh_paths_absolute: If mesh paths should be made absolute. This is needed for PyBullet to load the URDF file.
        """
        with open(path, mode="r") as f:
            urdf_string = self.fix_missing_inertial(f.read())
        urdf_string = self.remove_error_tags(urdf_string)
        urdf_string = self.fix_link_attributes(urdf_string)
        try:
            urdf_string = self.replace_relative_references_with_absolute_paths(urdf_string)
            urdf_string = self.fix_missing_inertial(urdf_string)
        except ResourceNotFound as e:
            logerr(f"Could not find resource package linked in this URDF")
            raise e
        urdf_string = self.make_mesh_paths_absolute(urdf_string, path) if make_mesh_paths_absolute else urdf_string
        self.write_description_to_file(urdf_string, save_path)

    def generate_from_parameter_server(self, name: str, save_path: str) -> None:
        urdf_string = get_parameter(name)
        urdf_string = self.replace_relative_references_with_absolute_paths(urdf_string)
        urdf_string = self.fix_missing_inertial(urdf_string)
        self.write_description_to_file(urdf_string, save_path)

    @property
    def joints(self) -> List[JointDescription]:
        """
        :return: A list of joints descriptions of this object.
        """
        if self._joints is None:
            self._init_joints()
        return self._joints

    def _init_joints(self) -> None:
        self._joints = [JointDescription(joint) for joint in self.parsed_description.joints]
        self._init_joints_map()

    @property
    def links(self) -> List[LinkDescription]:
        """
        :return: A list of link descriptions of this object.
        """
        if self._links is None:
            self._init_links()
        return self._links

    def _init_links(self) -> None:
        self._links = [LinkDescription(link) for link in self.parsed_description.links]
        self._init_links_map()

    def get_root(self) -> str:
        """
        :return: the name of the root link of this object.
        """
        return self.parsed_description.get_root()

    def get_tip(self) -> str:
        """
        :return: the name of the tip link of this object.
        :raises MultiplePossibleTipLinks: If there are multiple possible tip links.
        """
        link = self.get_root()
        while link in self.parsed_description.child_map:
            children = self.parsed_description.child_map[link]
            if len(children) > 1:
                # Multiple children, can't decide which one to take (e.g. fingers of a hand)
                raise MultiplePossibleTipLinks(self.parsed_description.name, link, [child[1] for child in children])
            else:
                child = children[0][1]
                link = child
        return link

    def get_chain(self, start_link_name: str, end_link_name: str, joints: Optional[bool] = True,
                  links: Optional[bool] = True, fixed: Optional[bool] = True) -> List[str]:
        """
        :param start_link_name: The name of the start link of the chain.
        :param end_link_name: The name of the end link of the chain.
        :param joints: Whether to include joints in the chain.
        :param links: Whether to include links in the chain.
        :param fixed: Whether to include fixed joints in the chain.
        :return: the chain of links from 'start_link_name' to 'end_link_name'.
        """
        return self.parsed_description.get_chain(start_link_name, end_link_name, joints, links, fixed)

    @staticmethod
    def replace_relative_references_with_absolute_paths(urdf_string: str) -> str:
        """
        Change paths for files in the URDF from ROS paths and file dir references to paths in the file system. Since
        World (PyBullet legacy) can't deal with ROS package paths.

        :param urdf_string: The name of the URDf on the parameter server
        :return: The URDF string with paths in the filesystem instead of ROS packages
        """
        new_urdf_string = ""
        for line in urdf_string.split('\n'):
            if "package://" in line:
                s = line.split('//')
                s1 = s[1].split('/')
                path = get_ros_package_path(s1[0])
                line = line.replace("package://" + s1[0], path)
            if 'file://' in line:
                line = line.replace("file://", './')
            new_urdf_string += line + '\n'

        return new_urdf_string

    @staticmethod
    def make_mesh_paths_absolute(urdf_string: str, urdf_file_path: str) -> str:
        """
        Convert all relative mesh paths in the URDF to absolute paths.

        :param urdf_string: The URDF description as string
        :param urdf_file_path: The path to the URDF file
        :returns: The new URDF description as string.
        """
        # Parse the URDF file
        root = ET.fromstring(urdf_string)

        # Iterate through all mesh tags
        for mesh in root.findall('.//mesh'):
            filename = mesh.attrib.get('filename', '')
            if filename:
                # If the filename is a relative path, convert it to an absolute path
                if not os.path.isabs(filename):
                    # Deduce the base path from the relative path
                    base_path = os.path.dirname(
                        os.path.abspath(os.path.join(os.path.dirname(urdf_file_path), filename)))
                    abs_path = os.path.abspath(os.path.join(base_path, os.path.basename(filename)))
                    mesh.set('filename', abs_path)

        return ET.tostring(root, encoding='unicode')

    @staticmethod
    def fix_missing_inertial(urdf_string: str) -> str:
        """
        Insert inertial tags for every URDF link that has no inertia.
        This is used to prevent Legacy(PyBullet) from dumping warnings in the terminal

        :param urdf_string: The URDF description as string
        :returns: The new, corrected URDF description as string.
        """

        inertia_tree = ET.ElementTree(ET.Element("inertial"))
        inertia_tree.getroot().append(ET.Element("mass", {"value": "0.1"}))
        inertia_tree.getroot().append(ET.Element("origin", {"rpy": "0 0 0", "xyz": "0 0 0"}))
        inertia_tree.getroot().append(ET.Element("inertia", {"ixx": "0.01",
                                                             "ixy": "0",
                                                             "ixz": "0",
                                                             "iyy": "0.01",
                                                             "iyz": "0",
                                                             "izz": "0.01"}))

        # create tree from string
        tree = ET.ElementTree(ET.fromstring(urdf_string))

        for link_element in tree.iter("link"):
            inertial = [*link_element.iter("inertial")]
            if len(inertial) == 0:
                link_element.append(inertia_tree.getroot())

        return ET.tostring(tree.getroot(), encoding='unicode')

    @staticmethod
    def remove_error_tags(urdf_string: str) -> str:
        """
        Remove all tags in the removing_tags list from the URDF since these tags are known to cause errors with the
        URDF_parser

        :param urdf_string: String of the URDF from which the tags should be removed
        :return: The URDF string with the tags removed
        """
        tree = ET.ElementTree(ET.fromstring(urdf_string))
        removing_tags = ["gazebo", "transmission"]
        for tag_name in removing_tags:
            all_tags = tree.findall(tag_name)
            for tag in all_tags:
                tree.getroot().remove(tag)

        return ET.tostring(tree.getroot(), encoding='unicode')

    @staticmethod
    def fix_link_attributes(urdf_string: str) -> str:
        """
        Remove the attribute 'type' from links since this is not parsable by the URDF parser.

        :param urdf_string: The string of the URDF from which the attributes should be removed
        :return: The URDF string with the attributes removed
        """
        tree = ET.ElementTree(ET.fromstring(urdf_string))

        for link in tree.iter("link"):
            if "type" in link.attrib.keys():
                del link.attrib["type"]

        return ET.tostring(tree.getroot(), encoding='unicode')

    @staticmethod
    def get_file_extension() -> str:
        """
        :return: The file extension of the URDF file.
        """
        return '.urdf'

    @property
    def origin(self) -> PoseStamped:
        return PoseStamped.from_list(self.parsed_description.origin.xyz,
                           quaternion_from_euler(*self.parsed_description.origin.rpy))

    @property
    def name(self) -> str:
        return self.parsed_description.name



from sqlalchemy.orm import Session
import tqdm
from typing_extensions import Optional
from .model import TaskTreeNode as TaskTreeNodeORM, self_mapped_classes, explicitly_mapped_classes
from ..language import LanguageNode
from ..plan import Plan


def insert(plan: Plan, session: Session, use_progress_bar: bool = True,
           progress_bar: Optional[tqdm.tqdm] = None):
    """
    Insert a plan into the database. Inserts all nodes and edges from the given plan into the database if the nodes and
    associated designators are present in the ORM model. Otherwise, check the genereate_orm.py file in scrip/ and the
    model.py file for explicitly mapped classes.

    :param plan: Plan to be inserted into the database.
    :param session: Session object for database interaction.
    :param use_progress_bar: If True, a progress bar will be displayed during the insertion process.
    :param progress_bar: Optional progress bar object. If not provided, a new one will be created.
    """

    if use_progress_bar:
        if not progress_bar:
            progress_bar = tqdm.tqdm(desc="Inserting Plan into database", leave=True, position=0, total=len(plan.nodes))

    # convert self to orm object
    for node in plan.nodes:
        if isinstance(node, LanguageNode):
            continue
        session.add(node)
        if use_progress_bar:
            progress_bar.update(1)

    session.commit()



import builtin_interfaces.msg
import rclpy
from . import node
import builtin_interfaces

class Time(builtin_interfaces.msg.Time):
    """
    Class to abstract the ROS2 Time, to make it more consistent with the ROS1 Time class.
    """
    def __init__(self, time=0, nsecs=0):
        super().__init__(sec=time)

    @classmethod
    def now(cls):
        return builtin_interfaces.msg.Time(**dict(zip(["sec", "nanosec"], node.get_clock().now().seconds_nanoseconds())))
        # return cls(*node.get_clock().now().seconds_nanoseconds())
        #return node.get_clock().now()

    def to_sec(self):
        return self.sec()[0]

# def Time(time=0.0):
#
#     return rclpy.time.Time(nanoseconds=time)

def Duration(duration=0.0):
    # return rclpy.duration.Duration(seconds=duration)
    return builtin_interfaces.msg.Duration(sec=duration)

def Rate(rate):
    return rate



class PinocchioConfig:
    error_threshold: float = 1e-4
    """
    The error threshold which needs to be reached for the goal to be considered as reached.
    """
    max_iterations: int = 10000
    """
    The maximum number of iterations for the inverse kinematics computation.
    """
    time_step: float = 1e-1
    """
    The time step for the integration of the joint configuration.
    """
    dampening_factor:float = 1e-12
    """
    The dampening factor for the inverse kinematics computation.
    """


from __future__ import annotations

import enum
from functools import lru_cache

import typing_extensions
from typing_extensions import List, Dict, Union, Tuple, TYPE_CHECKING, get_origin, get_args

from pycrap.ontologies import PhysicalObject
from .ros import logwarn

# Forward declaration of the class
HasParameters = object

if TYPE_CHECKING:
    LeafTypes = Union[float, int, str, bool, enum.Enum]

    ParameterDict = Dict[str, Union[LeafTypes, HasParameters]]

leaf_types = (int, float, str, bool, enum.Enum, type, PhysicalObject)


class HasParametersMeta(type):
    """
    Metaclass for flattening and reconstructing nested Python datastructures until they reach something mentioned in leaf_types.
    This is very similar to JAX PyTrees `https://docs.jax.dev/en/latest/pytrees.html#pytrees`_.
    Will scan the constructor of the class for fields to be used and not class variables.
    """

    _parameters: ParameterDict = None
    """
    A dictionary that maps field names to their types, including nested types.
    The keys are the names of the variables. The values are the types of the variables or nested types.
    """

    def __init__(cls, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        cls._parameters = {}
        cls.create_parameters(cls)

    @classmethod
    def create_parameters(cls, target_class):
        """
        Creates the flattened parameters for the given class.

        :param target_class: Class for which to create the parameters.
        """
        if not issubclass(target_class, HasParameters):
            raise TypeError(
                f"{target_class.__name__} must inherit from HasParameters. If you are using a dataclass use both the decorator has_parameters and inherit from HasParameters.")
        # calculate the field
        target_class._parameters = {}
        if hasattr(target_class, 'define_parameters'):
            try:
                target_class._parameters = target_class.define_parameters()
                return
            except NotImplementedError:
                pass

        for field_name, field_type in typing_extensions.get_type_hints(target_class.__init__).items():
            if field_name.startswith("_") or field_name == "return":
                continue
            # in case there are optional arguments
            if get_origin(field_type) is Union:
                type_a, type_b = get_args(field_type)
                field_type = type_a or type_b
                target_class._parameters[field_name] = field_type
                continue
            if get_origin(field_type) is type:
                type_type = get_args(field_type)[0]
                target_class._parameters[field_name] = type_type
                continue
            try:
                if issubclass(field_type, leaf_types) or issubclass(field_type, HasParameters):
                    target_class._parameters[field_name] = field_type
            except TypeError as e:
                logwarn(f"Filed type in {target_class.__name__} is not a leaf type: {field_type}")


class HasParameters(metaclass=HasParametersMeta):
    """
    Base class for everything that contains potentially parameters for a plan.
    """

    def flatten(self) -> List:
        """
        Flattens the object into a list of field values.

        :return: A list of flattened field values from the object.

        :raises TypeError: If the object is not an instance of the target class.
        """
        result = []
        for field_name, field_type in self._parameters.items():
            if issubclass(field_type, HasParameters):
                # sub_flattener = HasParameters(field_type.clazz)
                sub_obj = getattr(self, field_name)
                result.extend(sub_obj.flatten())
            else:
                result.append(getattr(self, field_name))
        return result

    @classmethod
    @lru_cache(maxsize=None)
    def flattened_parameters(cls) -> Dict[str, leaf_types]:
        """
        Returns a dictionary of all flattened fields and their types.

        :return: A dictionary mapping field names to their types.
        """
        flat_fields = {}
        for field_name, field_type in cls._parameters.items():
            if issubclass(field_type, HasParameters):
                sub_flat_fields = field_type.flattened_parameters()
                for sub_field_name, sub_field_type in sub_flat_fields.items():
                    flat_fields[f"{field_name}.{sub_field_name}"] = sub_field_type
            elif issubclass(field_type, leaf_types):
                flat_fields[field_name] = field_type
        return flat_fields

    @classmethod
    @lru_cache(maxsize=None)
    def number_of_fields(cls) -> int:
        """
        :return: The total number of fields in the flattened list.
        """
        return len(cls.flattened_parameters())

    @classmethod
    @lru_cache(maxsize=None)
    def field_indices(cls) -> Dict[str, Tuple[int, int]]:
        """
        :return: A dictionary mapping field names to their indices in the flattened list.
        """
        field_indices = {}
        index = 0
        for field_name, field_type in cls._parameters.items():
            if issubclass(field_type, HasParameters):
                field_indices[field_name] = (index, index + field_type.number_of_fields())
                index += field_type.number_of_fields()
            elif issubclass(field_type, leaf_types):
                field_indices[field_name] = (index, index + 1)
                index += 1
        return field_indices

    @classmethod
    def reconstruct(cls, flattened: List):
        """
        Reconstructs the object from a flattened list of field values.

        :param flattened: The flattened list of field values.

        :return: An instance of the target class with the reconstructed field values.

        :raises TypeError: If the object is not a list or if the length of the list does not match the number of fields.
        """
        if len(flattened) != len(cls.flattened_parameters()):
            raise ValueError("List length does not match number of flattened fields")

        kwargs = {}
        for field_name, field_type in cls._parameters.items():
            start, end = cls.field_indices()[field_name]
            if issubclass(field_type, HasParameters):
                sub_obj = flattened[start:end]
                kwargs[field_name] = field_type.reconstruct(sub_obj)
            else:
                kwargs[field_name] = flattened[start]

        return cls(**kwargs)

    @classmethod
    def define_parameters(cls) -> ParameterDict:
        """
        Override this method if you want to define a custom parameter dict for this class.

        :return: Dict like cls._parameters
        """
        raise NotImplementedError


T = typing_extensions.TypeVar("T")


def has_parameters(target_class: T) -> T:
    """
    Insert parameters of a class post construction.
    Use this when dataclasses should be combined with HasParameters.

    :param target_class: The class to get the parameters from.
    :return: The updated class
    """

    def wrapped_class():
        HasParametersMeta.create_parameters(target_class)
        return target_class

    return wrapped_class()



#!/usr/bin/env python3

"""Multiverse Client base class."""

from multiverse_client_pybind import MultiverseClientPybind  # noqa
from typing_extensions import Optional, List, Dict, Callable, TypeVar

from ...datastructures.dataclasses import MultiverseMetaData
from ...config.multiverse_conf import MultiverseConfig as Conf
from ...ros import loginfo, logwarn

T = TypeVar("T")


class MultiverseSocket:

    def __init__(
            self,
            port: str,
            host: str = Conf.HOST,
            meta_data: MultiverseMetaData = MultiverseMetaData(),
    ) -> None:
        """
        Initialize the MultiverseSocket, connect to the Multiverse Server and start the communication.

        :param port: The port of the client.
        :param host: The host of the client.
        :param meta_data: The metadata for the Multiverse Client as MultiverseMetaData.
        """
        if not isinstance(port, str) or port == "":
            raise ValueError(f"Must specify client port for {self.__class__.__name__}")
        self._send_data = None
        self.port = port
        self.host = host
        self._meta_data = meta_data
        self.client_name = self._meta_data.simulation_name
        self._multiverse_socket = MultiverseClientPybind(
            f"{Conf.SERVER_HOST}:{Conf.SERVER_PORT}"
        )
        self.request_meta_data = {
            "meta_data": self._meta_data.__dict__,
            "send": {},
            "receive": {},
        }
        self._api_callbacks: Optional[Dict] = None

        self._start_time = 0.0

    def run(self) -> None:
        """Run the client."""
        self.log_info("Start")
        self._run()

    def _run(self) -> None:
        """Run the client, should call the _connect_and_start() method. It's left to the user to implement this method
        in threaded or non-threaded fashion.
        """
        self._connect_and_start()

    def stop(self) -> None:
        """Stop the client."""
        self._disconnect()

    @property
    def request_meta_data(self) -> Dict:
        """The request_meta_data which is sent to the server.
        """
        return self._request_meta_data

    @request_meta_data.setter
    def request_meta_data(self, request_meta_data: Dict) -> None:
        """Set the request_meta_data, make sure to clear the `send` and `receive` field before setting the request
        """
        self._request_meta_data = request_meta_data
        self._multiverse_socket.set_request_meta_data(self._request_meta_data)

    @property
    def response_meta_data(self) -> Dict:
        """Get the response_meta_data.

        :return: The response_meta_data as a dictionary.
        """
        response_meta_data = self._multiverse_socket.get_response_meta_data()
        assert isinstance(response_meta_data, dict)
        if response_meta_data == {}:
            message = f"[Client {self.port}] Receive empty response meta data."
            self.log_warn(message)
        return response_meta_data

    def send_and_receive_meta_data(self):
        """
        Send and receive the metadata, this should be called before sending and receiving data.
        """
        self._communicate(True)

    def send_and_receive_data(self):
        """
        Send and receive the data, this should be called after sending and receiving the metadata.
        """
        self._communicate(False)

    @property
    def send_data(self) -> List[float]:
        """Get the send_data."""
        return self._send_data

    @send_data.setter
    def send_data(self, send_data: List[float]) -> None:
        """Set the send_data, the first element should be the current simulation time,
        the rest should be the data to send with the following order:
        double -> uint8_t -> uint16_t

        :param send_data: The data to send.
        """
        assert isinstance(send_data, list)
        self._send_data = send_data
        self._multiverse_socket.set_send_data(self._send_data)

    @property
    def receive_data(self) -> List[float]:
        """Get the receive_data, the first element should be the current simulation time,
        the rest should be the received data with the following order:
        double -> uint8_t -> uint16_t

        :return: The received data.
        """
        receive_data = self._multiverse_socket.get_receive_data()
        assert isinstance(receive_data, list)
        return receive_data

    @property
    def api_callbacks(self) -> Dict[str, Callable[[List[str]], List[str]]]:
        """Get the api_callbacks.

        :return: The api_callbacks as a dictionary of function names and their respective callbacks.
        """
        return self._api_callbacks

    @api_callbacks.setter
    def api_callbacks(self, api_callbacks: Dict[str, Callable[[List[str]], List[str]]]) -> None:
        """Set the api_callbacks.

        :param api_callbacks: The api_callbacks as a dictionary of function names and their respective callbacks.
        """
        self._multiverse_socket.set_api_callbacks(api_callbacks)
        self._api_callbacks = api_callbacks

    def _bind_request_meta_data(self, request_meta_data: T) -> T:
        """Bind the request_meta_data before sending it to the server.

        :param request_meta_data: The request_meta_data to bind.
        :return: The bound request_meta_data.
        """
        pass

    def _bind_response_meta_data(self, response_meta_data: T) -> T:
        """Bind the response_meta_data after receiving it from the server.

        :param response_meta_data: The response_meta_data to bind.
        :return: The bound response_meta_data.
        """
        pass

    def _bind_send_data(self, send_data: T) -> T:
        """Bind the send_data before sending it to the server.

        :param send_data: The send_data to bind.
        :return: The bound send_data.
        """
        pass

    def _bind_receive_data(self, receive_data: T) -> T:
        """Bind the receive_data after receiving it from the server.

        :param receive_data: The receive_data to bind.
        :return: The bound receive_data.
        """
        pass

    def _connect_and_start(self) -> None:
        """Connect to the server and start the client.
        """
        self._multiverse_socket.connect(self.host, self.port)
        self._multiverse_socket.start()
        self._start_time = self._multiverse_socket.get_time_now()

    def _disconnect(self) -> None:
        """Disconnect from the server.
        """
        self._multiverse_socket.disconnect()

    def _communicate(self, resend_request_meta_data: bool = False) -> bool:
        """Communicate with the server.

        :param resend_request_meta_data: Resend the request metadata.
        :return: True if the communication was successful, False otherwise.
        """
        return self._multiverse_socket.communicate(resend_request_meta_data)

    def _restart(self) -> None:
        """Restart the client.
        """
        self._disconnect()
        self._connect_and_start()

    def log_info(self, message: str) -> None:
        """Log information.

        :param message: The message to log.
        """
        loginfo(self._message_template(message))

    def log_warn(self, message: str) -> None:
        """Warn the user.

        :param message: The message to warn about.
        """
        logwarn(self._message_template(message))

    def _message_template(self, message: str) -> str:
        return (f"[{self.__class__.__name__}:{self.port}]: {message} : sim time {self.sim_time},"
                f" world time {self.world_time}")

    @property
    def world_time(self) -> float:
        """Get the world time from the server.

        :return: The world time.
        """
        return self._multiverse_socket.get_world_time()

    @property
    def sim_time(self) -> float:
        """Get the current simulation time.

        :return: The current simulation time.
        """
        return self._multiverse_socket.get_time_now() - self._start_time



import rospy

# Here so it can be imported from pycram.ros
from rospy import ServiceException

def Time(time: int = 0.0, nsecs: int = 0) -> rospy.Time:
    """
    Wrapper for rospy.Time to create a Time object.

    :param time: Time in seconds
    :param nsecs: Time in nanoseconds
    :return: Rospy Time object representing the given time
    """
    return rospy.Time(time, nsecs=nsecs)


def Duration(duration: float = 0.0) -> rospy.Duration:
    """
    Wrapper for rospy.Duration to create a Duration object.

    :param duration: Duration in seconds
    :return: A rospy Duration object representing the given duration
    """
    return rospy.Duration.from_sec(duration)


def Rate(rate: float) -> rospy.Rate:
    """
    Wrapper for rospy.Rate to create a Rate object.

    :param rate: Rate in Hz
    :return: A rospy Rate object representing the given rate
    """
    return rospy.Rate(rate)



from __future__ import annotations

import typing_extensions
from typing_extensions import Type, Optional

from pycrap.urdf_parser import parse_furniture
from pycrap.ontologies import Base, PhysicalObject

if typing_extensions.TYPE_CHECKING:
    from .world import World


class HasConcept:
    """
    A mixin class that adds an ontological concept and individual to a class that will be registered in PyCRAP.
    """

    ontology_concept: Type[Base] = Base
    """
    The ontological concept that this class represents.
    """

    ontology_individual: Optional[Base] = None
    """
    The individual in the ontology that is connected with this class.
    """

    onto_name: Optional[str] = None
    """
    The name of the individual in the ontology.
    """

    def __init__(self, world: World, name: Optional[str] = None, concept: Type[Base] = PhysicalObject,
                 parse_name: bool = True):

        if world.is_prospection_world:
            return
        self.onto_name = name
        self.ontology_concept = concept
        self.ontology_individual = self.ontology_concept(name=name.lower() if name is not None else None,
                                                         namespace=world.ontology.ontology)
        if parse_name:
            inferred_concept = parse_furniture(name)
            if inferred_concept is not None:
                self.ontology_individual.is_a = [inferred_concept]

        world.ontology.python_objects[self.ontology_individual] = self



def get_service_proxy(topic_name: str, service_message) -> None:
    """
    Get a service proxy for a given topic name and service message type.

    :param topic_name: The name of the service.
    :param service_message: The type of the service message.
    :return: A service proxy for the specified topic and message type.
    """
    raise NotImplementedError("ROS is not available. Cannot get service proxy.")

def wait_for_service(topic_name: str) -> None:
    """
    Wait for a service to become available.

    :param topic_name: The name of the service.
    """
    raise NotImplementedError("ROS is not available. Cannot wait for service.")



from dataclasses import dataclass
from pycrap.ontologies import PhysicalObject, Location
from .object_designator import ObjectDesignatorDescription, ObjectPart
from ..datastructures.enums import MovementType,WaypointsMovementType
from ..failure_handling import try_motion
from ..failures import PerceptionObjectNotFound, ToolPoseNotReachedError
from ..object_descriptors.urdf import LinkDescription, ObjectDescription
from ..plan import with_plan
from ..process_module import ProcessModuleManager
from ..datastructures.enums import ObjectType, Arms, GripperState, ExecutionType, DetectionTechnique, DetectionState

from typing_extensions import Dict, Optional, Type, List
from ..datastructures.pose import PoseStamped
from ..designator import BaseMotion
from ..world_concepts.world_object import Object
from ..external_interfaces.robokudo import robokudo_found


@with_plan
@dataclass
class MoveMotion(BaseMotion):
    """
    Moves the robot to a designated location
    """

    target: PoseStamped
    """
    Location to which the robot should be moved
    """

    keep_joint_states: bool = False
    """
    Keep the joint states of the robot during/at the end of the motion
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.navigate().execute(self)

@with_plan
@dataclass
class MoveTCPMotion(BaseMotion):
    """
    Moves the Tool center point (TCP) of the robot
    """

    target: PoseStamped
    """
    Target pose to which the TCP should be moved
    """
    arm: Arms
    """
    Arm with the TCP that should be moved to the target
    """
    allow_gripper_collision: Optional[bool] = None
    """
    If the gripper can collide with something
    """
    movement_type: Optional[MovementType] = MovementType.CARTESIAN
    """
    The type of movement that should be performed.
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        try_motion(pm_manager.move_tcp(), self, ToolPoseNotReachedError)


    def __str__(self):
        return (f"MoveTCPMotion:\n"
                f"Target: {self.target}\n"
                f"Arm: {self.arm}\n"
                f"AllowGripperCollision: {self.allow_gripper_collision}\n"
                f"MovementType: {self.movement_type}")

    def __repr__(self):
        return self.__str__()

@with_plan
@dataclass
class LookingMotion(BaseMotion):
    """
    Lets the robot look at a point
    """
    target: PoseStamped

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.looking().execute(self)


@with_plan
@dataclass
class MoveGripperMotion(BaseMotion):
    """
    Opens or closes the gripper
    """

    motion: GripperState
    """
    Motion that should be performed, either 'open' or 'close'
    """
    gripper: Arms
    """
    Name of the gripper that should be moved
    """
    allow_gripper_collision: Optional[bool] = None
    """
    If the gripper is allowed to collide with something
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.move_gripper().execute(self)


    def __str__(self):
        return (f"MoveGripperMotion:\n"
                f"Motion: {self.motion}\n"
                f"Gripper: {self.gripper}\n"
                f"AllowGripperCollision: {self.allow_gripper_collision}")

    def __repr__(self):
        return self.__str__()

@with_plan
@dataclass
class DetectingMotion(BaseMotion):
    """
    Tries to detect an object in the FOV of the robot

    returns: ObjectDesignatorDescription.Object or Error: PerceptionObjectNotFound
    """

    technique: DetectionTechnique
    """
    Detection technique that should be used
    """
    state: DetectionState
    """
    State of the detection
    """
    object_designator_description: Optional[Object] = None
    """
    Description of the object that should be detected
    """
    region: Optional[Location] = None
    """
    Region in which the object should be detected
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        obj_dict = pm_manager.detecting().execute(self)
        return obj_dict


@with_plan
@dataclass
class MoveArmJointsMotion(BaseMotion):
    """
    Moves the joints of each arm into the given position
    """

    left_arm_poses: Optional[Dict[str, float]] = None
    """
    Target positions for the left arm joints
    """
    right_arm_poses: Optional[Dict[str, float]] = None
    """
    Target positions for the right arm joints
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.move_arm_joints().execute(self)

@with_plan
@dataclass
class WorldStateDetectingMotion(BaseMotion):
    """
    Detects an object based on the world state.
    """

    object_type: ObjectType
    """
    Object type that should be detected
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.world_state_detecting().execute(self)

@with_plan
@dataclass
class MoveJointsMotion(BaseMotion):
    """
    Moves any joint on the robot
    """

    names: list
    """
    List of joint names that should be moved 
    """
    positions: list
    """
    Target positions of joints, should correspond to the list of names
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.move_joints().execute(self)

@with_plan
@dataclass
class OpeningMotion(BaseMotion):
    """
    Designator for opening container
    """

    object_part: ObjectDescription.Link
    """
    Object designator for the drawer handle
    """
    arm: Arms
    """
    Arm that should be used
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.open().execute(self)

@with_plan
@dataclass
class ClosingMotion(BaseMotion):
    """
    Designator for closing a container
    """

    object_part: ObjectDescription.Link
    """
    Object designator for the drawer handle
    """
    arm: Arms
    """
    Arm that should be used
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.close().execute(self)

@with_plan
@dataclass
class TalkingMotion(BaseMotion):
    """
    Talking Motion, lets the robot say a sentence.
    """

    cmd: str
    """
    Talking Motion, let the robot say a sentence.
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        return pm_manager.talk().execute(self)


@with_plan
@dataclass
class MoveTCPWaypointsMotion(BaseMotion):
    """
    Moves the Tool center point (TCP) of the robot
    """

    waypoints: List[PoseStamped]
    """
    Waypoints the TCP should move along 
    """
    arm: Arms
    """
    Arm with the TCP that should be moved to the target
    """
    allow_gripper_collision: Optional[bool] = None
    """
    If the gripper can collide with something
    """
    movement_type: WaypointsMovementType = WaypointsMovementType.ENFORCE_ORIENTATION_FINAL_POINT
    """
    The type of movement that should be performed.
    """

    def perform(self):
        pm_manager = ProcessModuleManager.get_manager()
        pm_manager.move_tcp_waypoints().execute(self)

    def __str__(self):
        return (f"MoveTCPWaypointsMotion:\n"
                f"Waypoints: {self.waypoints}\n"
                f"Arm: {self.arm}\n"
                f"AllowGripperCollision: {self.allow_gripper_collision}\n"
                f"MovementType: {self.movement_type}")

    def __repr__(self):
        return self.__str__()


import numpy as np
from pycram.tf_transformations import quaternion_from_euler
from typing_extensions import Optional, List

from pycram.testing import BulletWorldTestCase
from pycram.datastructures.enums import JointType
from pycram.datastructures.pose import PoseStamped
from pycram.robot_description import RobotDescription
from pycram.validation.error_checkers import PoseErrorChecker, PositionErrorChecker, \
    OrientationErrorChecker, RevoluteJointPositionErrorChecker, PrismaticJointPositionErrorChecker, \
    MultiJointPositionErrorChecker
from pycram.validation.goal_validator import GoalValidator, PoseGoalValidator, \
    PositionGoalValidator, OrientationGoalValidator, JointPositionGoalValidator, MultiJointPositionGoalValidator, \
    MultiPoseGoalValidator, MultiPositionGoalValidator, MultiOrientationGoalValidator


class TestGoalValidator(BulletWorldTestCase):

    def test_single_pose_goal(self):
        pose_goal_validators = PoseGoalValidator(self.milk.get_pose)
        self.validate_pose_goal(pose_goal_validators)

    def test_single_pose_goal_generic(self):
        pose_goal_validators = GoalValidator(PoseErrorChecker(), self.milk.get_pose)
        self.validate_pose_goal(pose_goal_validators)

    def validate_pose_goal(self, goal_validator):
        milk_goal_pose = PoseStamped.from_list([1.3, 1.5, 0.9])
        goal_validator.register_goal(milk_goal_pose)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertAlmostEqual(goal_validator.current_error.tolist()[0], 0.5, places=5)
        self.assertAlmostEqual(goal_validator.current_error.tolist()[1], 0, places=5)
        self.milk.set_pose(milk_goal_pose)
        self.assertEqual(self.milk.get_pose(), milk_goal_pose)
        self.assertTrue(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 1)
        self.assertAlmostEqual(goal_validator.current_error.tolist()[0], 0, places=5)
        self.assertAlmostEqual(goal_validator.current_error.tolist()[1], 0, places=5)

    def test_single_position_goal_generic(self):
        goal_validator = GoalValidator(PositionErrorChecker(), self.cereal.get_position_as_list)
        self.validate_position_goal(goal_validator)

    def test_single_position_goal(self):
        goal_validator = PositionGoalValidator(self.cereal.get_position_as_list)
        self.validate_position_goal(goal_validator)

    def validate_position_goal(self, goal_validator):
        cereal_goal_position = [1.3, 1.5, 0.95]
        goal_validator.register_goal(cereal_goal_position)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertEqual(goal_validator.current_error, 0.8)
        self.cereal.set_position(cereal_goal_position)
        self.assertEqual(self.cereal.get_position_as_list(), cereal_goal_position)
        self.assertTrue(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 1)
        self.assertEqual(goal_validator.current_error, 0)

    def test_single_orientation_goal_generic(self):
        goal_validator = GoalValidator(OrientationErrorChecker(), self.cereal.get_orientation_as_list)
        self.validate_orientation_goal(goal_validator)

    def test_single_orientation_goal(self):
        goal_validator = OrientationGoalValidator(self.cereal.get_orientation_as_list)
        self.validate_orientation_goal(goal_validator)

    def validate_orientation_goal(self, goal_validator):
        cereal_goal_orientation = quaternion_from_euler(0, 0, np.pi / 2)
        goal_validator.register_goal(cereal_goal_orientation)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertEqual(goal_validator.current_error, [np.pi / 2])
        self.cereal.set_orientation(cereal_goal_orientation)
        for v1, v2 in zip(self.cereal.get_orientation_as_list(), cereal_goal_orientation):
            self.assertAlmostEqual(v1, v2, places=5)
        self.assertTrue(goal_validator.goal_achieved)
        self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, 1, places=5)
        self.assertAlmostEqual(goal_validator.current_error.tolist()[0], 0, places=5)

    def test_single_revolute_joint_position_goal_generic(self):
        goal_validator = GoalValidator(RevoluteJointPositionErrorChecker(), self.robot.get_joint_position)
        self.validate_revolute_joint_position_goal(goal_validator)

    def test_single_revolute_joint_position_goal(self):
        goal_validator = JointPositionGoalValidator(self.robot.get_joint_position)
        self.validate_revolute_joint_position_goal(goal_validator, JointType.REVOLUTE)

    def validate_revolute_joint_position_goal(self, goal_validator, joint_type: Optional[JointType] = None):
        goal_joint_position = -np.pi / 8
        joint_name = 'l_shoulder_lift_joint'
        if joint_type is not None:
            goal_validator.register_goal(goal_joint_position, joint_type, joint_name)
        else:
            goal_validator.register_goal(goal_joint_position, joint_name)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertEqual(goal_validator.current_error, abs(goal_joint_position))

        for percent in [0.5, 1]:
            self.robot.set_joint_position('l_shoulder_lift_joint', goal_joint_position * percent)
            self.assertEqual(self.robot.get_joint_position('l_shoulder_lift_joint'), goal_joint_position * percent)
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], abs(goal_joint_position) * (1 - percent),
                                   places=5)

    def test_single_prismatic_joint_position_goal_generic(self):
        goal_validator = GoalValidator(PrismaticJointPositionErrorChecker(), self.robot.get_joint_position)
        self.validate_prismatic_joint_position_goal(goal_validator)

    def test_single_prismatic_joint_position_goal(self):
        goal_validator = JointPositionGoalValidator(self.robot.get_joint_position)
        self.validate_prismatic_joint_position_goal(goal_validator, JointType.PRISMATIC)

    def validate_prismatic_joint_position_goal(self, goal_validator, joint_type: Optional[JointType] = None):
        goal_joint_position = 0.2
        torso = RobotDescription.current_robot_description.torso_joint
        if joint_type is not None:
            goal_validator.register_goal(goal_joint_position, joint_type, torso)
        else:
            goal_validator.register_goal(goal_joint_position, torso)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertEqual(goal_validator.current_error, abs(goal_joint_position))

        for percent in [0.5, 1]:
            self.robot.set_joint_position(torso, goal_joint_position * percent)
            self.assertEqual(self.robot.get_joint_position(torso), goal_joint_position * percent)
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], abs(goal_joint_position) * (1 - percent),
                                   places=5)

    def test_multi_joint_goal_generic(self):
        joint_types = [JointType.PRISMATIC, JointType.REVOLUTE]
        goal_validator = GoalValidator(MultiJointPositionErrorChecker(joint_types),
                                       lambda x: list(self.robot.get_multiple_joint_positions(x).values()))
        self.validate_multi_joint_goal(goal_validator)

    def test_multi_joint_goal(self):
        joint_types = [JointType.PRISMATIC, JointType.REVOLUTE]
        goal_validator = MultiJointPositionGoalValidator(
            lambda x: list(self.robot.get_multiple_joint_positions(x).values()))
        self.validate_multi_joint_goal(goal_validator, joint_types)

    def validate_multi_joint_goal(self, goal_validator, joint_types: Optional[List[JointType]] = None):
        goal_joint_positions = np.array([0.2, -np.pi / 4])
        joint_names = ['torso_lift_joint', 'l_shoulder_lift_joint']
        if joint_types is not None:
            goal_validator.register_goal(goal_joint_positions, joint_types, joint_names)
        else:
            goal_validator.register_goal(goal_joint_positions, joint_names)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertTrue(np.allclose(goal_validator.current_error, np.array([0.2, abs(-np.pi / 4)]), atol=0.001))

        for percent in [0.5, 1]:
            current_joint_positions = goal_joint_positions * percent
            self.robot.set_multiple_joint_positions(dict(zip(joint_names, current_joint_positions.tolist())))
            self.assertTrue(np.allclose(self.robot.get_joint_position('torso_lift_joint'), current_joint_positions[0],
                                        atol=0.001))
            self.assertTrue(
                np.allclose(self.robot.get_joint_position('l_shoulder_lift_joint'), current_joint_positions[1],
                            atol=0.001))
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], abs(0.2) * (1 - percent), places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[1], abs(-np.pi / 4) * (1 - percent), places=5)

    def test_list_of_poses_goal_generic(self):
        goal_validator = GoalValidator(PoseErrorChecker(is_iterable=True),
                                       lambda: [self.robot.get_pose(), self.robot.get_pose()])
        self.validate_list_of_poses_goal(goal_validator)

    def test_list_of_poses_goal(self):
        goal_validator = MultiPoseGoalValidator(lambda: [self.robot.get_pose(), self.robot.get_pose()])
        self.validate_list_of_poses_goal(goal_validator)

    def validate_list_of_poses_goal(self, goal_validator):
        position_goal = [0.0, 1.0, 0.0]
        orientation_goal = np.array([0, 0, np.pi / 2])
        poses_goal = [PoseStamped.from_list(position_goal, quaternion_from_euler(*orientation_goal.tolist())),
                      PoseStamped.from_list(position_goal, quaternion_from_euler(*orientation_goal.tolist()))]
        goal_validator.register_goal(poses_goal)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertTrue(
            np.allclose(goal_validator.current_error, np.array([1.0, np.pi / 2, 1.0, np.pi / 2]), atol=0.001))

        for percent in [0.5, 1]:
            current_orientation_goal = orientation_goal * percent
            current_pose_goal = PoseStamped.from_list([0.0, 1.0 * percent, 0.0],
                                            quaternion_from_euler(*current_orientation_goal.tolist()))
            self.robot.set_pose(current_pose_goal)
            self.assertTrue(np.allclose(self.robot.get_position_as_list(), current_pose_goal.position.to_list(),
                                        atol=0.001))
            self.assertTrue(np.allclose(self.robot.get_orientation_as_list(), current_pose_goal.orientation.to_list(),
                                        atol=0.001))
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], 1 - percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[1], np.pi * (1 - percent) / 2, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[2], (1 - percent), places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[3], np.pi * (1 - percent) / 2, places=5)

    def test_list_of_positions_goal_generic(self):
        goal_validator = GoalValidator(PositionErrorChecker(is_iterable=True),
                                       lambda: [self.robot.get_position_as_list(), self.robot.get_position_as_list()])
        self.validate_list_of_positions_goal(goal_validator)

    def test_list_of_positions_goal(self):
        goal_validator = MultiPositionGoalValidator(lambda: [self.robot.get_position_as_list(),
                                                             self.robot.get_position_as_list()])
        self.validate_list_of_positions_goal(goal_validator)

    def validate_list_of_positions_goal(self, goal_validator):
        position_goal = [0.0, 1.0, 0.0]
        positions_goal = [position_goal, position_goal]
        goal_validator.register_goal(positions_goal)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertTrue(np.allclose(goal_validator.current_error, np.array([1.0, 1.0]), atol=0.001))

        for percent in [0.5, 1]:
            current_position_goal = [0.0, 1.0 * percent, 0.0]
            self.robot.set_position(current_position_goal)
            self.assertTrue(np.allclose(self.robot.get_position_as_list(), current_position_goal, atol=0.001))
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], 1 - percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[1], 1 - percent, places=5)

    def test_list_of_orientations_goal_generic(self):
        goal_validator = GoalValidator(OrientationErrorChecker(is_iterable=True),
                                       lambda: [self.robot.get_orientation_as_list(),
                                                self.robot.get_orientation_as_list()])
        self.validate_list_of_orientations_goal(goal_validator)

    def test_list_of_orientations_goal(self):
        goal_validator = MultiOrientationGoalValidator(lambda: [self.robot.get_orientation_as_list(),
                                                                self.robot.get_orientation_as_list()])
        self.validate_list_of_orientations_goal(goal_validator)

    def validate_list_of_orientations_goal(self, goal_validator):
        orientation_goal = np.array([0, 0, np.pi / 2])
        orientations_goals = [quaternion_from_euler(*orientation_goal.tolist()),
                              quaternion_from_euler(*orientation_goal.tolist())]
        goal_validator.register_goal(orientations_goals)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertTrue(np.allclose(goal_validator.current_error, np.array([np.pi / 2, np.pi / 2]), atol=0.001))

        for percent in [0.5, 1]:
            current_orientation_goal = orientation_goal * percent
            self.robot.set_orientation(quaternion_from_euler(*current_orientation_goal.tolist()))
            self.assertTrue(np.allclose(self.robot.get_orientation_as_list(),
                                        quaternion_from_euler(*current_orientation_goal.tolist()),
                                        atol=0.001))
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], np.pi * (1 - percent) / 2, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[1], np.pi * (1 - percent) / 2, places=5)

    def test_list_of_revolute_joint_positions_goal_generic(self):
        goal_validator = GoalValidator(RevoluteJointPositionErrorChecker(is_iterable=True),
                                       lambda x: list(self.robot.get_multiple_joint_positions(x).values()))
        self.validate_list_of_revolute_joint_positions_goal(goal_validator)

    def test_list_of_revolute_joint_positions_goal(self):
        goal_validator = MultiJointPositionGoalValidator(
            lambda x: list(self.robot.get_multiple_joint_positions(x).values()))
        self.validate_list_of_revolute_joint_positions_goal(goal_validator, [JointType.REVOLUTE, JointType.REVOLUTE])

    def validate_list_of_revolute_joint_positions_goal(self, goal_validator,
                                                       joint_types: Optional[List[JointType]] = None):
        goal_joint_position = -np.pi / 4
        goal_joint_positions = np.array([goal_joint_position, goal_joint_position])
        joint_names = ['l_shoulder_lift_joint', 'r_shoulder_lift_joint']
        if joint_types is not None:
            goal_validator.register_goal(goal_joint_positions, joint_types, joint_names)
        else:
            goal_validator.register_goal(goal_joint_positions, joint_names)
        self.assertFalse(goal_validator.goal_achieved)
        self.assertEqual(goal_validator.actual_percentage_of_goal_achieved, 0)
        self.assertTrue(np.allclose(goal_validator.current_error,
                                    np.array([abs(goal_joint_position), abs(goal_joint_position)]), atol=0.001))

        for percent in [0.5, 1]:
            current_joint_position = goal_joint_positions * percent
            self.robot.set_multiple_joint_positions(dict(zip(joint_names, current_joint_position)))
            self.assertTrue(np.allclose(list(self.robot.get_multiple_joint_positions(joint_names).values()),
                                        current_joint_position, atol=0.001))
            if percent == 1:
                self.assertTrue(goal_validator.goal_achieved)
            else:
                self.assertFalse(goal_validator.goal_achieved)
            self.assertAlmostEqual(goal_validator.actual_percentage_of_goal_achieved, percent, places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[0], abs(goal_joint_position) * (1 - percent),
                                   places=5)
            self.assertAlmostEqual(goal_validator.current_error.tolist()[1], abs(goal_joint_position) * (1 - percent),
                                   places=5)



from ..ros import  get_ros_package_path
from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager
from ..datastructures.enums import Arms, Grasp, GripperState, GripperType, StaticJointState
from ..units import meter

filename = get_ros_package_path('pycram') + '/resources/robots/' + "ur5_robotiq" + '.urdf'

ur5_description = RobotDescription("ur5_robotiq", "world", "base_link", "ee_link",
                                   filename)

################################## Arm ##################################
arm = KinematicChainDescription("manipulator", "base_link", "wrist_3_link", ur5_description.urdf_object, arm_type=Arms.RIGHT)

arm.add_static_joint_states(StaticJointState.Park, {'shoulder_pan_joint': 0.0,
                                     'shoulder_lift_joint': 0.0,
                                     'elbow_joint': 0.0,
                                     'wrist_1_joint': 0.0,
                                     'wrist_2_joint': 0.0,
                                     'wrist_3_joint': 0.0})

ur5_description.add_kinematic_chain_description(arm)

################################## Gripper ##################################
gripper = EndEffectorDescription("gripper", "robotiq_85_base_link", "robotiq_85_right_finger_link",
                                 ur5_description.urdf_object)

gripper.add_static_joint_states(GripperState.OPEN, {'robotiq_85_left_finger_joint': 0.0,
                                         'robotiq_85_right_finger_joint': 0.0,
                                         'robotiq_85_left_inner_knuckle_joint': 0.0,
                                         'robotiq_85_right_inner_knuckle_joint': 0.0,
                                         'robotiq_85_left_finger_tip_joint': 0.0,
                                         'robotiq_85_right_finger_tip_joint': 0.0})
gripper.add_static_joint_states(GripperState.CLOSE, {'robotiq_85_left_finger_joint': 1,
                                          'robotiq_85_right_finger_joint': 1,
                                          'robotiq_85_left_inner_knuckle_joint': 1.0,
                                          'robotiq_85_right_inner_knuckle_joint': 1.0,
                                          'robotiq_85_left_finger_tip_joint': 1.0,
                                          'robotiq_85_right_finger_tip_joint': 1.0})
gripper.end_effector_type = GripperType.PARALLEL
gripper.opening_distance = 0.085 * meter
arm.end_effector = gripper

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(ur5_description)



from ..datastructures.dataclasses import ManipulatorData
from ..helper import get_robot_urdf_and_mjcf_file_paths, find_multiverse_resources_path
from ..robot_description import RobotDescriptionManager, create_manipulator_description
from ..ros import  logwarn, loginfo

data = ManipulatorData(
    name="ur5e",
    relative_dir="universal_robot",
    base_link="base_link",

    arm_end_link="wrist_3_link",
    joint_names=['shoulder_pan_joint',
                 'shoulder_lift_joint',
                 'elbow_joint',
                 'wrist_1_joint',
                 'wrist_2_joint',
                 'wrist_3_joint'],
    home_joint_values=[3.14, -1.56, 1.58, -1.57, -1.57, 0.0],

    gripper_name="gripper-2F-85",
    gripper_relative_dir="robotiq",
    gripper_tool_frame="right_pad",

    gripper_joint_names=['right_driver_joint',
                         'right_coupler_joint',
                         'right_spring_link_joint',
                         'right_follower_joint',
                         'left_driver_joint',
                         'left_coupler_joint',
                         'left_spring_link_joint',
                         'left_follower_joint'],
    closed_joint_values=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    open_joint_values=[0.798, 0.00366, 0.796, -0.793, 0.798, 0.00366, 0.796, -0.793],
    opening_distance=None,  # TODO: Check this value

    gripper_cmd_topic="/gripper_command",
    gripper_open_cmd_value=0.0,
    gripper_close_cmd_value=255.0)

multiverse_resources = find_multiverse_resources_path()

urdf_filename, mjcf_filename = None, None
if multiverse_resources is not None:
    urdf_filename, mjcf_filename = get_robot_urdf_and_mjcf_file_paths(data.name, data.relative_dir,
                                                                      multiverse_resources=multiverse_resources)

if mjcf_filename is None or urdf_filename is None:
    loginfo(f"Could not initialize {data.name} description as Multiverse resources path not found.")
else:
    robot_description = create_manipulator_description(data, urdf_filename, mjcf_filename)

    # Add to RobotDescriptionManager
    rdm = RobotDescriptionManager()
    rdm.register_description(robot_description)



import sys
import unittest

from pycram.testing import BulletWorldTestCase
from pycram.ros import set_logger_level, logwarn, logerr, logdebug
from pycram.datastructures.enums import LoggerLevel


class TestLogging(BulletWorldTestCase):
    """Testcase for the pycram logging functions."""

    @unittest.skipIf(sys.version_info.minor > 9, "ROS logging does not work in higher python versions")
    def test_set_logger_level(self):
        set_logger_level(LoggerLevel.DEBUG)
        logdebug("This is a debug message, it should be printed")
        logwarn("This is a warning, it should be printed")
        logerr("This is an error, it should be printed")
        set_logger_level(LoggerLevel.ERROR)
        logdebug("This is a debug message, it should not be printed")
        logwarn("This is a warning, it should not be printed")
        logerr("This is an error, it should be printed")



from ..ros import  get_ros_package_path

from ..datastructures.dataclasses import VirtualMobileBaseJoints
from ..datastructures.enums import GripperState, Arms, Grasp, TorsoState, GripperType, StaticJointState
from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..helper import get_robot_description_path
from ..units import meter

filename = get_ros_package_path('pycram') + '/resources/robots/' + "tiago_dual" + '.urdf'

mjcf_filename = get_robot_description_path("pal_robotics", "tiago_dual")

tiago_description = RobotDescription("tiago_dual", "base_link", "torso_lift_link", "torso_lift_joint",
                                     filename,
                                     virtual_mobile_base_joints=VirtualMobileBaseJoints(),
                                     mjcf_path=mjcf_filename)

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left_arm", "torso_lift_link", "arm_left_7_link",
                                     tiago_description.urdf_object, arm_type=Arms.LEFT)

left_arm.add_static_joint_states(StaticJointState.Park, {'arm_left_1_joint': 0.27,
                                          'arm_left_2_joint': -1.07,
                                          'arm_left_3_joint': 1.5,
                                          'arm_left_4_joint': 1.96,
                                          'arm_left_5_joint': -2.0,
                                          'arm_left_6_joint': 1.2,
                                          'arm_left_7_joint': 0.5})

tiago_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################
left_gripper = EndEffectorDescription("left_gripper", "gripper_left_link", "gripper_left_grasping_frame",
                                      tiago_description.urdf_object)

left_gripper.add_static_joint_states(GripperState.OPEN, {'gripper_left_left_finger_joint': 0.048,
                                                         'gripper_left_right_finger_joint': 0.048})

left_gripper.add_static_joint_states(GripperState.CLOSE, {'gripper_left_left_finger_joint': 0.0,
                                                          'gripper_left_right_finger_joint': 0.0})
left_gripper.end_effector_type = GripperType.PARALLEL
left_gripper.opening_distance = 0.09 * meter  # measured
left_arm.end_effector = left_gripper

################################## Right Arm ##################################
right_arm = KinematicChainDescription("right_arm", "torso_lift_link", "arm_right_7_link",
                                      tiago_description.urdf_object, arm_type=Arms.RIGHT)

right_arm.add_static_joint_states(StaticJointState.Park, {'arm_right_1_joint': 0.27,
                                           'arm_right_2_joint': -1.07,
                                           'arm_right_3_joint': 1.5,
                                           'arm_right_4_joint': 2.0,
                                           'arm_right_5_joint': -2.0,
                                           'arm_right_6_joint': 1.2,
                                           'arm_right_7_joint': 0.5})

tiago_description.add_kinematic_chain_description(right_arm)

################################## Right Gripper ##################################
right_gripper = EndEffectorDescription("right_gripper", "gripper_right_link", "gripper_right_grasping_frame",
                                       tiago_description.urdf_object)

right_gripper.add_static_joint_states(GripperState.OPEN, {'gripper_right_left_finger_joint': 0.048,
                                                          'gripper_right_right_finger_joint': 0.048})

right_gripper.add_static_joint_states(GripperState.CLOSE, {'gripper_right_left_finger_joint': 0.0,
                                                           'gripper_right_right_finger_joint': 0.0})
right_gripper.end_effector_type = GripperType.PARALLEL
right_gripper.opening_distance = 0.09 * meter  # measured
right_arm.end_effector = right_gripper

################################## Torso ##################################
torso = KinematicChainDescription("torso", "torso_fixed_link", "torso_lift_link",
                                  tiago_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"torso_lift_joint": 0.3})

torso.add_static_joint_states(TorsoState.MID, {"torso_lift_joint": 0.15})

torso.add_static_joint_states(TorsoState.LOW, {"torso_lift_joint": 0})

tiago_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
camera = CameraDescription("xtion_optical_frame", "xtion_optical_frame", 0.99483, 0.75049, 1.0665, 1.4165)
tiago_description.add_camera_description(camera)

################################## Neck ##################################
tiago_description.add_kinematic_chain("neck", "torso_lift_link", "head_2_link")
tiago_description.set_neck(yaw_joint="head_1_joint", pitch_joint="head_2_joint")

################################# Grasps ##################################
front_grasp = [0, 0, 0, 1]
right_gripper.update_all_grasp_orientations(front_grasp)
left_gripper.update_all_grasp_orientations(front_grasp)


# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(tiago_description)



import math
import unittest

from pycram.datastructures.pose import PoseStamped, TransformStamped, Quaternion, Vector3, AxisIdentifier


class TestPose(unittest.TestCase):

    def test_pose_creation(self):
        p = PoseStamped.from_list([1, 2, 3], [0, 0, 0, 1], "test")

        self.assertEqual(p.position.to_list(), [1, 2, 3])
        self.assertEqual(p.orientation.to_list(), [0, 0, 0, 1])
        self.assertEqual(p.frame_id, "test")
        self.assertEqual(p.pose.to_list(), [[1, 2, 3], [0, 0, 0, 1]])

    def test_pose_to_transform(self):
        p = PoseStamped.from_list([3, 2, 1], [0, 0, 1, 0], "map")

        transform = p.to_transform_stamped("test_frame")

        self.assertEqual(transform, TransformStamped.from_list([3, 2, 1], [0, 0, 1, 0], "map", "test_frame"))

    def test_pose_edit(self):
        p = PoseStamped.from_list([3, 4, 5], [0, 1, 0, 0], "map")

        p.position = Vector3(1, 1, 1)
        self.assertEqual(p.position.to_list(), [1, 1, 1])
        p.position.x = 2
        self.assertEqual(p.position.to_list(), [2, 1, 1])
        p.position = Vector3(3, 3, 3)
        self.assertEqual(p.position.to_list(), [3, 3, 3])

        p.orientation = Quaternion(0, 0, 0, 1)
        self.assertEqual(p.orientation.to_list(), [0, 0, 0, 1])
        p.orientation.x = 1
        self.assertEqual(p.orientation.to_list(), [1, 0, 0, 1])
        p.orientation = Quaternion(0, 0, 1, 0)
        self.assertEqual(p.orientation.to_list(), [0, 0, 1, 0])

    def test_pose_copy(self):
        p1 = PoseStamped.from_list([1, 2, 3], [0, 0, 0, 1], "map")
        p2 = p1.copy()

        self.assertEqual(p1, p2)
        self.assertFalse(p1 is p2)

    def test_transform_creation(self):
        t = TransformStamped.from_list([1, 2, 3], [0, 0, 0, 1], "map", "test_frame")

        self.assertEqual(t.translation.to_list(), [1, 2, 3])
        self.assertEqual(t.rotation.to_list(), [0, 0, 0, 1])
        self.assertEqual(t.frame_id, "map")
        self.assertEqual(t.child_frame_id, "test_frame")

    def test_transform_edit(self):
        t = TransformStamped.from_list([3, 2, 1], [0, 1, 0, 0], "map", "test_frame")

        t.translation = Vector3(2, 2, 2)
        self.assertEqual(t.translation.to_list(), [2, 2, 2])
        t.translation.x = 3
        self.assertEqual(t.translation.to_list(), [3, 2, 2])
        t.translation = Vector3(1, 1, 1)
        self.assertEqual(t.translation.to_list(), [1, 1, 1])

        t.rotation = Quaternion(1, 0, 0, 0)
        self.assertEqual(t.rotation.to_list(), [1, 0, 0, 0])
        t.rotation.y = 1
        self.assertEqual(t.rotation.to_list(), [1, 1, 0, 0])
        t.rotation = Quaternion(0, 0, 0, 1)
        self.assertEqual(t.rotation.to_list(), [0, 0, 0, 1])

    def test_transform_copy(self):
        t = TransformStamped.from_list([1, 1, 1], [0, 0, 0, 1], "map", "test_frame")

        t_copy = t.copy()
        self.assertEqual(t, t_copy)
        self.assertFalse(t is t_copy)

    def test_transform_multiplication(self):
        t = TransformStamped.from_list([1, 2, 3], [0, 0, 0, 1], "map", "test_frame")
        t2 = TransformStamped.from_list([3, 2, 1], [0, 0, 0, 1], "test_frame", "final_frame")

        mul_t = t * t2

        self.assertEqual(mul_t.translation.to_list(), [4, 4, 4])

    def test_is_facing_2d_axis(self):
        a = PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1], "map")  # facing +x
        b = PoseStamped.from_list([1, 0, 0], [0, 0, 0, 1], "map")

        facing, angle = a.is_facing_2d_axis(b, axis=AxisIdentifier.X)
        self.assertTrue(facing)
        self.assertAlmostEqual(angle, 0, delta=1e-6)

        # now test Y alignment (should be 90 deg difference)
        facing_y, angle_y = a.is_facing_2d_axis(b, axis=AxisIdentifier.Y)
        self.assertFalse(facing_y)
        self.assertAlmostEqual(abs(angle_y), math.pi / 2, delta=1e-6)

    def test_is_facing_x_or_y(self):
        a = PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1], "map")
        b = PoseStamped.from_list([1, 0, 0], [0, 0, 0, 1], "map")

        self.assertTrue(a.is_facing_x_or_y(b))

        # reverse direction
        b.position.x = -1
        self.assertFalse(a.is_facing_x_or_y(b))